# coding=utf-8

import os

import geometry as geo
import numpy as np
from comptests import comptest, get_comptests_output_dir, run_module_tests
from numpy.testing import assert_almost_equal

from duckietown_world import LaneSegment, PlacedObject, RectangularArea, SE2Transform
from duckietown_world.rules import evaluate_rules
from duckietown_world.rules.rule import make_timeseries
from duckietown_world.seqs.tsequence import SampledSequence
from duckietown_world.svg_drawing import draw_static
from duckietown_world.world_duckietown.differential_drive_dynamics import (
    DifferentialDriveDynamicsParameters,
    WheelVelocityCommands,
)
from duckietown_world.world_duckietown.duckiebot import DB18
from duckietown_world.world_duckietown.lane_segment import get_distance_two
from duckietown_world.world_duckietown.map_loading import load_map
from duckietown_world.world_duckietown.tile_template import load_tile_types


def same_point(a, b):
    assert_almost_equal(a.p, b.p)
    assert_almost_equal(a.theta, b.theta)


@comptest
def lane_pose2():
    templates = load_tile_types()

    for name, ls in templates.items():
        if not isinstance(ls, LaneSegment):
            continue

        print(ls.get_lane_lengths())
        length = ls.get_lane_length()

        # first pose is the first control point
        lp0 = ls.lane_pose(along_lane=0.0, relative_heading=0.0, lateral=0.0)
        t0 = ls.SE2Transform_from_lane_pose(lp0)

        p0 = ls.control_points[0]
        same_point(p0, t0)

        # at 1 pose it is the second control point
        lp1 = ls.lane_pose(along_lane=length, relative_heading=0.0, lateral=0.0)
        t1 = ls.SE2Transform_from_lane_pose(lp1)

        p1 = ls.control_points[-1]
        same_point(p1, t1)

        l1 = length * 0.2
        l2 = length * 0.3
        d = l2 - l1


@comptest
def lane_pose3():
    templates = load_tile_types()

    for name, ls in templates.items():
        if not isinstance(ls, LaneSegment):
            continue

        length = ls.get_lane_length()

        # first pose is the first control point
        l1 = length * 0.3
        l2 = length * 0.4
        d = l2 - l1

        lp1 = ls.lane_pose(along_lane=l1, relative_heading=0.0, lateral=0.0)
        lp2 = ls.lane_pose(along_lane=l2, relative_heading=0.0, lateral=0.0)

        q1 = ls.SE2Transform_from_lane_pose(lp1).as_SE2()
        q2 = ls.SE2Transform_from_lane_pose(lp2).as_SE2()

        d2 = get_distance_two(q1, q2)

        assert_almost_equal(d, d2, decimal=3)


@comptest
def lane_pose4():
    templates = load_tile_types()

    for name, ls in templates.items():
        if not isinstance(ls, LaneSegment):
            continue
        print(name)

        lp1 = ls.lane_pose_random()
        # print('lp1: %s' % lp1)
        q1 = ls.SE2Transform_from_lane_pose(lp1)
        # print('q1: %s' % q1)
        lp2 = ls.lane_pose_from_SE2Transform(q1, tol=0.001)

        # print('lp2: %s' % lp2)
        assert_almost_equal(lp1.along_lane, lp2.along_lane, decimal=3)
        assert_almost_equal(lp1.lateral, lp2.lateral, decimal=3)
        assert_almost_equal(lp1.relative_heading, lp2.relative_heading, decimal=3)


@comptest
def center_point1():
    outdir = get_comptests_output_dir()
    templates = load_tile_types()

    for k, v in templates.items():
        if isinstance(v, LaneSegment):

            area = RectangularArea((-2, -2), (3, 3))
            dest = os.path.join(outdir, k)

            N = len(v.control_points)
            betas = list(np.linspace(-2, N + 1, 20))
            betas.extend(range(N))
            betas = sorted(betas)
            transforms = []
            for timestamp in betas:
                beta = timestamp
                p = v.center_point(beta)
                # print('%s: %s' % (beta, geo.SE2.friendly(p)))

                transform = SE2Transform.from_SE2(p)
                transforms.append(transform)

            c = SampledSequence[SE2Transform](betas, transforms)
            v.set_object("a", PlacedObject(), ground_truth=c)
            draw_static(v, dest, area=area)


@comptest
def lane_pose_test1():
    outdir = get_comptests_output_dir()

    # load one of the maps (see the list using dt-world-draw-maps)
    dw = load_map("udem1")

    v = 5

    # define a SampledSequence with timestamp, command
    commands_sequence = SampledSequence.from_iterator(
        [
            # we set these velocities at 1.0
            (1.0, WheelVelocityCommands(0.1 * v, 0.1 * v)),
            # at 2.0 we switch and so on
            (2.0, WheelVelocityCommands(0.1 * v, 0.4 * v)),
            (4.0, WheelVelocityCommands(0.1 * v, 0.4 * v)),
            (5.0, WheelVelocityCommands(0.1 * v, 0.2 * v)),
            (6.0, WheelVelocityCommands(0.1 * v, 0.1 * v)),
        ]
    )

    # we upsample the sequence by 5
    commands_sequence = commands_sequence.upsample(5)

    ## Simulate the dynamics of the vehicle
    # start from q0
    q0 = geo.SE2_from_translation_angle([1.8, 0.7], 0)

    # instantiate the class that represents the dynamics
    dynamics = reasonable_duckiebot()
    # this function integrates the dynamics
    poses_sequence = get_robot_trajectory(dynamics, q0, commands_sequence)

    #################
    # Visualization and rule evaluation

    # Creates an object 'duckiebot'
    ego_name = "duckiebot"
    db = DB18()  # class that gives the appearance

    # convert from SE2 to SE2Transform representation
    transforms_sequence = poses_sequence.transform_values(SE2Transform.from_SE2)
    # puts the object in the world with a certain "ground_truth" constraint
    dw.set_object(ego_name, db, ground_truth=transforms_sequence)

    # Rule evaluation (do not touch)
    interval = SampledSequence.from_iterator(enumerate(commands_sequence.timestamps))
    evaluated = evaluate_rules(
        poses_sequence=transforms_sequence, interval=interval, world=dw, ego_name=ego_name,
    )
    timeseries = make_timeseries(evaluated)
    # Drawing
    area = RectangularArea((0, 0), (3, 3))
    draw_static(dw, outdir, area=area, timeseries=timeseries)


def integrate_commands(s0, commands_sequence):
    states = [s0]
    timestamps = commands_sequence.timestamps
    t0 = timestamps[0]
    yield t0, s0
    for i in range(len(timestamps) - 1):
        dt = timestamps[i + 1] - timestamps[i]
        commands = commands_sequence.values[i]
        s_prev = states[-1]
        s_next = s_prev.integrate(dt, commands)
        states.append(s_next)
        t = timestamps[i + 1]
        yield t, s_next


def get_robot_trajectory(factory, q0, commands_sequence):
    assert isinstance(commands_sequence, SampledSequence)
    t0 = commands_sequence.timestamps[0]

    # initialize trajectory
    c0 = q0, geo.se2.zero()
    s0 = factory.initialize(c0=c0, t0=t0)

    states_sequence = SampledSequence.from_iterator(integrate_commands(s0, commands_sequence))
    f = lambda _: _.TSE2_from_state()[0]
    poses_sequence = states_sequence.transform_values(f, np.ndarray)
    return poses_sequence


def reasonable_duckiebot():
    radius = 0.1
    radius_left = radius
    radius_right = radius
    wheel_distance = 0.5
    dddp = DifferentialDriveDynamicsParameters(
        radius_left=radius_left, radius_right=radius_right, wheel_distance=wheel_distance,
    )
    return dddp


if __name__ == "__main__":
    run_module_tests()



# coding=utf-8
import numpy as np
from contracts import contract

import geometry as geo
from duckietown_serialization_ds1 import Serializable
from .platform_dynamics import PlatformDynamicsFactory, PlatformDynamics

__all__ = ["Integrator2D"]


class Integrator2D(PlatformDynamicsFactory, PlatformDynamics, Serializable):
    """
        This represents the dynamics of a 2D integrator.

        Commands are sequences of 2 numbers / array of two numbers.

    """

    @classmethod
    @contract(c0="TSE2")
    def initialize(cls, c0, t0=0, seed=None):
        """
            This class initializes the dynamics at a given configuration
        """
        # pose, velocity in SE(2), se(2)
        q, v = c0
        # get position p from pose
        p, theta = geo.translation_angle_from_SE2(q)
        # get linear velocity from se(2)
        v2d, _ = geo.linear_angular_from_se2(v)
        # create the integrator2d initial state
        return Integrator2D(p, v2d, t0)

    @contract(p0="seq[2]")
    def __init__(self, p0, v0, t0):
        """
        :param p0: initial point
        :param v0: initial velocity
        :param t0: initial time
        """
        self.t0 = t0
        self.v0 = v0
        self.p0 = p0

    def integrate(self, dt, commands):
        """
            Returns the next state after applying commands for time dt.

            :param dt: for how long to apply the commands
            :param commands: sequences of two numbers
            :return: another Integrator2D
        """
        # convert things to float, array
        dt = float(dt)
        commands = np.array(commands, np.float64)

        # time incremensts by dt
        t1 = self.t0 + dt
        # we integrate commands
        p1 = self.p0 + dt * commands
        # the velocity is the commands
        v1 = commands
        return Integrator2D(p1, v1, t1)

    @contract(returns="TSE2")
    def TSE2_from_state(self):
        """
            For visualization purposes, this function gets a configuration in SE2
            from the internal state.
        """
        # pose
        q = geo.SE2_from_R2(self.p0)
        # velocity
        linear = self.v0
        angular = 0.0
        v = geo.se2_from_linear_angular(linear, angular)
        return q, v



# coding=utf-8

from duckietown_world import logger
from duckietown_world.svg_drawing.misc import mime_from_fn, draw_axes

from ..geo import PlacedObject

__all__ = [
    "GenericObject",
    "Duckie",
    "Decoration",
    "Sign",
    "Sign4WayIntersect",
    "SignLeftTIntersect",
    "SignRightTIntersect",
    "SignStop",
    "SignTIntersect",
    "SignTLightAhead",
    "SignOneWayRight",
    "SignOneWayLeft",
    "SignDuckCrossing",
    "SignYield",
    "SignNoLeftTurn",
    "SignNoRightTurn",
    "SignDoNotEnter",
    "SignParking",
    "SignPedestrian",
    "Bus",
    "House",
    "Tree",
    "Truck",
    "Cone",
    "Vehicle",
    "Barrier",
    "Building",
    "SIGNS",
    "SIGNS_ALIASES",
    "get_canonical_sign_name",
]


class GenericObject(PlacedObject):
    def __init__(self, kind, **kwargs):
        self.kind = kind
        PlacedObject.__init__(self, **kwargs)

    def params_to_json_dict(self):
        return dict(kind=self.kind)

    def draw_svg(self, drawing, g):
        c = drawing.circle(center=(0, 0), r=0.1)
        g.add(c)


class Duckie(PlacedObject):
    def draw_svg(self, drawing, g):
        c = drawing.circle(center=(0, 0), r=0.065, fill="yellow", stroke="black", stroke_width=0.01)
        g.add(c)


class Decoration(PlacedObject):
    pass


class Tree(Decoration):
    def draw_svg(self, drawing, g):
        c = drawing.circle(center=(0, 0), r=0.25, fill="green", stroke="black", stroke_width=0.01)
        g.add(c)


class Vehicle(PlacedObject):
    pass


class Cone(PlacedObject):
    def draw_svg(self, drawing, g):
        c = drawing.circle(center=(0, 0), r=0.03, fill="orange", stroke="black", stroke_width=0.01)
        g.add(c)


class Bus(Vehicle):
    def draw_svg(self, drawing, g):
        L, W = 0.3, 0.2
        c = drawing.rect(
            insert=(-L / 2, -W / 2), fill="grey", size=(L, W), stroke_width="0.01", stroke="#eeeeee",
        )
        g.add(c)


class Truck(Vehicle):
    def draw_svg(self, drawing, g):
        L, W = 0.4, 0.2
        c = drawing.rect(
            insert=(-L / 2, -W / 2), fill="blue", size=(L, W), stroke_width="0.01", stroke="#eeeeee",
        )
        g.add(c)


class House(Decoration):
    def draw_svg(self, drawing, g):
        L, W = 0.5, 0.25
        c = drawing.rect(
            insert=(-L / 2, -W / 2), fill="red", size=(L, W), stroke_width="0.01", stroke="#eeeeee",
        )
        g.add(c)


class Barrier(Decoration):
    def draw_svg(self, drawing, g):
        L, W = 0.5, 0.25
        c = drawing.rect(
            insert=(-L / 2, -W / 2), fill="pink", size=(L, W), stroke_width="0.01", stroke="#eeeeee",
        )
        g.add(c)


class Building(Decoration):
    # TODO
    def draw_svg(self, drawing, g):
        L, W = 0.5, 0.25
        c = drawing.rect(
            insert=(-L / 2, -W / 2), fill="red", size=(L, W), stroke_width="0.01", stroke="#eeeeee",
        )
        g.add(c)


class Sign(PlacedObject):
    def __init__(self, tag=None, **kwargs):
        PlacedObject.__init__(self, **kwargs)
        self.tag = tag

    def draw_svg(self, drawing, g):
        texture = self.get_name_texture()
        # x = -0.2
        CM = 0.01
        PAPER_WIDTH, PAPER_HEIGHT = 8.5 * CM, 15.5 * CM
        PAPER_THICK = 0.01

        BASE_SIGN = 5 * CM
        WIDTH_SIGN = 1.1 * CM

        y = -1.5 * PAPER_HEIGHT  # XXX not sure why this is negative
        y = BASE_SIGN
        x = -PAPER_WIDTH / 2
        try:
            from .map_loading import get_texture_file

            fn = get_texture_file(texture)
        except KeyError:
            msg = "Cannot find texture for %s" % texture
            logger.warning(msg)

            c = drawing.rect(
                insert=(x, y),
                fill="white",
                size=(PAPER_WIDTH, PAPER_HEIGHT),
                stroke_width="0.01",
                stroke="black",
            )
            g.add(c)

        else:
            with open(fn, "rb") as _:
                texture = _.read()
            from duckietown_world.world_duckietown.tile import data_encoded_for_src

            href = data_encoded_for_src(texture, mime_from_fn(fn))

            img = drawing.image(
                href=href,
                size=(PAPER_WIDTH, PAPER_HEIGHT),
                insert=(x, y),
                stroke_width=0.001,
                stroke="black",
                style="transform: rotate(90deg) rotate(90deg) scaleX(-1)  rotate(-90deg);"
                "border: solid 1px black ",
            )
            img.attribs["class"] = "sign-paper"
            g.add(img)

        c = drawing.rect(
            insert=(-BASE_SIGN / 2, -BASE_SIGN / 2),
            fill="lightgreen",
            stroke_width=0.001,
            stroke="black",
            size=(BASE_SIGN, BASE_SIGN),
        )
        g.add(c)
        draw_axes(drawing, g)
        alpha = 1.2
        c = drawing.rect(
            insert=(-WIDTH_SIGN / 2, -BASE_SIGN / 2 * alpha),
            fill="#FFCB9C",
            stroke_width=0.001,
            stroke="black",
            size=(WIDTH_SIGN, BASE_SIGN * alpha),
        )
        g.add(c)

        c = drawing.rect(
            insert=(+0.01 + -PAPER_THICK / 2, -PAPER_WIDTH / 2),
            fill="white",
            stroke_width=0.001,
            stroke="black",
            size=(PAPER_THICK, PAPER_WIDTH),
        )
        g.add(c)

    def get_name_texture(self):
        for k, v in SIGNS.items():
            if v.__name__ == type(self).__name__:
                return k

        raise KeyError(type(self).__name__)


class SignStop(Sign):
    pass


class SignLeftTIntersect(Sign):
    pass


class SignRightTIntersect(Sign):
    pass


class SignTIntersect(Sign):
    pass


class Sign4WayIntersect(Sign):
    pass


class SignTLightAhead(Sign):
    pass


class SignOneWayRight(Sign):
    pass


class SignOneWayLeft(Sign):
    pass


class SignDuckCrossing(Sign):
    pass


class SignYield(Sign):
    pass


class SignNoLeftTurn(Sign):
    pass


class SignNoRightTurn(Sign):
    pass


class SignDoNotEnter(Sign):
    pass


class SignParking(Sign):
    pass


class SignPedestrian(Sign):
    pass


SIGNS = {
    "sign_left_T_intersect": SignLeftTIntersect,
    "sign_right_T_intersect": SignRightTIntersect,
    "sign_T_intersect": SignTIntersect,
    "sign_4_way_intersect": Sign4WayIntersect,
    "sign_t_light_ahead": SignTLightAhead,
    "sign_stop": SignStop,
    "sign_1_way_right": SignOneWayRight,
    "sign_1_way_left": SignOneWayLeft,
    "sign_duck_crossing": SignDuckCrossing,
    "sign_yield": SignYield,
    "sign_parking": SignParking,
    "sign_pedestrian": SignPedestrian,
    "sign_do_not_enter": SignDoNotEnter,
    "sign_no_left_turn": SignNoLeftTurn,
    "sign_no_right_turn": SignNoRightTurn,
}

SIGNS_ALIASES = {
    "T-intersection": "sign_T_intersect",
    "oneway-right": "sign_1_way_right",
    "oneway-left": "sign_1_way_left",
    "duck-crossing": "sign_duck_crossing",
    "stop": "sign_stop",
    "yield": "sign_yield",
    "no-left-turn": "sign_no_left_turn",
    "t-light-ahead": "sign_t_light_ahead",
    "pedestrian": "sign_pedestrian",
    "no-right-turn": "sign_no_right_turn",
    "parking": "sign_parking",
    "right-T-intersect": "sign_right_T_intersect",
    "left-T-intersect": "sign_left_T_intersect",
    "4-way-intersect": "sign_4_way_intersect",
    "do-not-enter": "sign_do_not_enter",
}


def get_canonical_sign_name(sign_name):
    if sign_name in SIGNS:
        return sign_name
    if sign_name in SIGNS_ALIASES:
        return SIGNS_ALIASES[sign_name]
    raise KeyError(sign_name)



# coding=utf-8
import typing
from abc import abstractmethod

# from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    ClassVar,
    Iterator,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)

from zuper_commons.types import ZValueError
from zuper_typing import dataclass, Generic

__all__ = [
    "GenericSequence",
    "UndefinedAtTime",
    "SampledSequence",
    "IterateDT",
    "iterate_with_dt",
    'Timestamp',
]


class UndefinedAtTime(Exception):
    pass


X = TypeVar("X")
Y = TypeVar("Y")
Timestamp = float


@dataclass
class GenericSequence(Generic[X]):
    CONTINUOUS = "continuous-sampling"

    @abstractmethod
    def at(self, t: Timestamp) -> Generic:
        """ Raises UndefinedAtTime if not defined at t. """

    @abstractmethod
    def get_start(self) -> Optional[Timestamp]:
        """ Returns the timestamp for start, or None if -infinity. """

    @abstractmethod
    def get_end(self) -> Optional[Timestamp]:
        """ Returns the timestamp for start, or None if +infinity. """

    @abstractmethod
    def get_sampling_points(self) -> Union[str, typing.Sequence[Timestamp]]:
        """
            Returns the lists of interesting points.

            Returns the special value CONTINUOUS if
            dense sampling is possible
        """


Sequence = GenericSequence


@dataclass
class SampledSequence(GenericSequence):
    """ A sampled time sequence. Only defined at certain points. """

    timestamps: List[Timestamp]
    values: List[X]

    XT: ClassVar[Type[X]] = object

    def __post_init__(self):
        values = list(self.values)
        timestamps = list(self.timestamps)

        if len(timestamps) != len(values):
            msg = "Length mismatch"
            raise ValueError(msg)

        for t in timestamps:
            if not isinstance(t, (float, int)):
                msg = "I expected a number, got %s" % type(t)
                raise ValueError(msg)
        for i in range(len(timestamps) - 1):
            dt = timestamps[i + 1] - timestamps[i]
            if dt <= 0:
                msg = "Invalid dt = %s at i = %s; ts= %s" % (dt, i, timestamps)
                raise ValueError(msg)
        timestamps = list(map(Timestamp, timestamps))
        self.timestamps = timestamps
        self.values = values

    def at(self, t: Timestamp) -> X:
        try:
            i = self.timestamps.index(t)
        except ValueError:
            msg = "Could not find timestamp %s in %s" % (t, self.timestamps)
            raise UndefinedAtTime(msg)
        else:
            return self.values[i]

    def at_or_previous(self, t: Timestamp) -> X:
        try:
            return self.at(t)
        except UndefinedAtTime:
            pass

        # last_t = self.timestamps[0]
        last_i = 0
        for i in range(len(self.timestamps)):
            if self.timestamps[i] < t:
                # last_t = self.timestamps[i]
                last_i = i
            else:
                break
        return self.values[last_i]

    def get_sampling_points(self) -> List[Timestamp]:
        return list(self.timestamps)

    def get_start(self) -> Timestamp:
        if not self.timestamps:
            msg = "Empty sequence"
            raise ValueError(msg)
        return self.timestamps[0]

    def get_end(self) -> Timestamp:
        if not self.timestamps:
            msg = "Empty sequence"
            raise ValueError(msg)
        return self.timestamps[-1]

    def __iter__(self):
        return zip(self.timestamps, self.values).__iter__()

    @classmethod
    def from_iterator(cls, i: Iterator[X], T: Type[X] = object) -> "SampledSequence[X]":
        timestamps = []
        values = []
        for t, v in i:
            assert isinstance(t, (float, int)), type(t)
            t = Timestamp(t)
            timestamps.append(t)
            values.append(v)
        return SampledSequence[T](timestamps, values)

    def __len__(self) -> int:
        return len(self.timestamps)

    def transform_values(self, f: Callable[[X], Y], YT: Type[Y] = object) -> "SampledSequence[Y]":
        values = []
        timestamps = []
        for t, _ in self:
            res = f(_)
            if res is not None:
                values.append(res)
                timestamps.append(t)

        return SampledSequence[YT](timestamps, values)

    def upsample(self, n: int) -> "SampledSequence[X]":
        timestamps = []
        values = []
        for i in range(len(self.timestamps) - 1):
            for k in range(n):
                t0 = self.timestamps[i]
                t1 = self.timestamps[i + 1]
                t = t0 + (k * 1.0 / n) * (t1 - t0)
                timestamps.append(Timestamp(t))
                values.append(self.values[i])
        timestamps.append(self.timestamps[-1])
        values.append(self.values[-1])
        return SampledSequence[self.XT](timestamps, values)


def downsample(s: SampledSequence[X], M: int) -> SampledSequence[X]:
    timestamps = []
    values = []
    for i in range(len(s.timestamps) - 1):
        if i % M == 0:
            timestamps.append(s.timestamps[i])
            values.append(s.values[i])
    return SampledSequence[s.XT](timestamps, values)


@dataclass
class IterateDT(Generic[Y]):
    t0: Timestamp
    t1: Timestamp
    dt: float
    v0: Y
    v1: Y


def iterate_with_dt(sequence: SampledSequence[X]) -> Iterator[IterateDT[X]]:
    """ yields t0, t1, dt, v0, v1 """
    timestamps = sequence.timestamps
    values = sequence.values
    for i in range(len(timestamps) - 1):
        t0 = timestamps[i]
        assert isinstance(t0, float), type(t0)
        t1 = timestamps[i + 1]
        v0 = values[i]
        v1 = values[i + 1]
        dt = float(t1 - t0)
        X = type(sequence).XT
        yield IterateDT[X](t0, t1, dt, v0, v1)  # XXX


@dataclass
class SampledSequenceBuilder(Generic[X]):
    timestamps: List[Timestamp] = None
    values: List[X] = None

    XT: ClassVar[Type[X]] = Any

    def __post_init__(self):
        self.timestamps = self.timestamps or []
        self.values = self.values or []

    def add(self, t: Timestamp, v: X):
        # self.timestamps = self.timestamps or []
        # self.values = self.values or []

        # print(type(self), self.__dict__)
        if self.timestamps:
            if t == self.timestamps[-1]:
                msg = "Repeated time stamp"
                raise ZValueError(msg, t=t, timestamps=self.timestamps)
        self.timestamps.append(t)
        self.values.append(v)

    def __len__(self) -> int:
        return len(self.timestamps)

    def as_sequence(self) -> SampledSequence:
        return SampledSequence[self.XT](timestamps=self.timestamps, values=self.values)



# coding=utf-8
from comptests import comptest, get_comptests_output_dir, run_module_tests
from duckietown_world import draw_static
from duckietown_world.world_duckietown.sampling_poses import sample_good_starting_pose
import duckietown_world as dw


@comptest
def test_pose_sampling_1():

    m = dw.load_map("4way")

    along_lane = 0.2
    only_straight = True
    for i in range(45):
        q = sample_good_starting_pose(m, only_straight=only_straight, along_lane=along_lane)
        ground_truth = dw.SE2Transform.from_SE2(q)
        m.set_object(f"sampled-{i}", dw.DB18(), ground_truth=ground_truth)

    outdir = get_comptests_output_dir()
    draw_static(m, outdir)


if __name__ == "__main__":
    run_module_tests()



# coding=utf-8

import numpy as np
from duckietown_serialization_ds1 import Serializable

import geometry
from duckietown_world.geo import Matrix2D, Transform

__all__ = ["TileCoords", "TileRelativeTransform"]

ALLOWED_ORIENTATIONS = ["S", "E", "N", "W"]


class TileCoords(Transform, Serializable):
    i: int
    j: int
    orientation: str  # N S W E

    def asmatrix2d(self):
        angle0 = {"N": 0, "E": -np.pi / 2, "S": np.pi, "W": +np.pi / 2}[self.orientation] + np.pi / 2

        angle = {"N": np.pi / 2, "E": 0, "S": np.pi + np.pi / 2, "W": np.pi}[self.orientation]

        assert np.allclose(angle, angle0)

        p = [self.i + 0.5, self.j + 0.5]
        M = geometry.SE2_from_translation_angle(p, angle)
        return Matrix2D(M)

    def __init__(self, i: int, j: int, orientation: str):
        self.i = i
        self.j = j
        self.orientation = orientation

    def params_to_json_dict(self):
        return dict(i=self.i, j=self.j, orientation=self.orientation)


class TileRelativeTransform(Transform, Serializable):
    def __init__(self, p, z, theta):
        self.p = np.array(p, dtype="float64")
        self.z = z
        self.theta = float(theta)

    def asmatrix2d(self):
        M = geometry.SE2_from_translation_angle(self.p, self.theta)
        return Matrix2D(M)

    def params_to_json_dict(self):
        return dict(p=self.p, z=self.z, theta=self.theta)



# coding=utf-8
from duckietown_world.geo import PlacedObject
from duckietown_world.svg_drawing import draw_axes
from .other_objects import Vehicle

__all__ = ["Duckiebot", "DB18"]


class Duckiebot(Vehicle):
    def __init__(self, width, length, height, *args, color: str = None, **kwargs):
        PlacedObject.__init__(self, *args, **kwargs)

        self.width = width
        self.height = height
        self.length = length
        self.color = color or "red"

    def draw_svg(self, drawing, g):
        L, W = self.length, self.width
        rect = drawing.rect(
            insert=(-L * 0.5, -W * 0.5),
            size=(L, W),
            fill=self.color,
            # style='opacity:0.4',
            stroke_width="0.01",
            stroke="darkred",
        )
        rect.width = "0.1em"
        g.add(rect)

        draw_axes(drawing, g)
        # print(g.tostring())


class DB18(Duckiebot):
    def __init__(self, *args, **kwargs):
        width = 0.13  # + 0.02
        length = 0.18
        height = 0.12
        Duckiebot.__init__(self, width=width, length=length, height=height, *args, **kwargs)



# coding=utf-8
from typing import Tuple

from duckietown_world.world_duckietown.tile_coords import ALLOWED_ORIENTATIONS
from ..geo import PlacedObject
from ..world_duckietown import Tile, TileCoords

__all__ = ["TileMap"]


def tilename_from_ij(i: int, j: int):
    tile_name = "tile-%d-%d" % (i, j)
    return tile_name


def ij_from_tilename(tilename: str) -> Tuple[int, int]:
    tokens = tilename.split("-")
    i, j = int(tokens[1]), int(tokens[2])
    return i, j


class TileMap(PlacedObject):
    H: int  # i
    W: int  # j

    def __init__(self, H: int, W: int, **kwargs):
        PlacedObject.__init__(self, **kwargs)
        self.H = H
        self.W = W

    def params_to_json_dict(self):
        return dict(H=self.H, W=self.W)

    def __getitem__(self, coords: Tuple[int, int]) -> Tile:
        tilename = tilename_from_ij(coords[0], coords[1])
        if not tilename in self.children:
            msg = 'Tile "%s" not available' % coords.__repr__()
            raise KeyError(msg)
        return self.children[coords]

    def add_tile(self, i: int, j: int, orientation: str, tile: Tile, can_be_outside: bool = False):
        if not can_be_outside:
            assert 0 <= i < self.H, (i, self.H)
            assert 0 <= j < self.W, (j, self.W)

        assert orientation in ALLOWED_ORIENTATIONS, orientation
        tile_name = "tile-%d-%d" % (i, j)
        placement = TileCoords(i, j, orientation)
        self.set_object(tile_name, tile, ground_truth=placement)

    def get_drawing_children(self):
        def key(x):
            if x.startswith("tile"):
                return 0, x
            else:
                return 1, x

        return sorted(self.children, key=key)



# coding=utf-8
import yaml

from duckietown_serialization_ds1 import Serializable


from duckietown_world.utils.memoizing import memoized_reset

# language=yaml
data = """

# lane = 22cm
# tile = 58.5
# lane_rel = 22/58.5
go_right: &go_right
    ~LaneSegment:
      width: &width 0.376
      control_points: 
        - ~SE2Transform:
            p: [-0.50, -0.22]
            theta_deg: 0
        - ~SE2Transform:
            p: [-0.30, -0.30]
            theta_deg: -45
        - ~SE2Transform:
            p: [-0.22, -0.50]
            theta_deg: -90

go_straight: &go_straight
    ~LaneSegment:
      width: *width
      control_points: 
        - ~SE2Transform:
            p: [-0.5, -0.22]
            theta_deg: 0
        - ~SE2Transform:
            p: [+0.5, -0.22]
            theta_deg: 0

go_left: &go_left
    ~LaneSegment:
      width: *width
      control_points: 
        - ~SE2Transform:
            p: [-0.5, -0.22]
            theta_deg: 0 

        - ~SE2Transform:
            p: [0.0, 0.0]
            theta_deg: 45
 
        - ~SE2Transform:
            p: [+0.22, +0.50]
            theta_deg: 90 

            
straight:
    ~PlacedObject:
        children:
            lane1: *go_straight
            lane2: *go_straight
        spatial_relations:
            lane1: {~SE2Transform:}
            lane2: {~SE2Transform: {theta_deg: 180}}

curve_right: &curve_right
    ~PlacedObject:
        children:
            lane1: *go_right
            lane2: *go_left
        spatial_relations:
            lane1: {~SE2Transform:}
            lane2: {~SE2Transform: {theta_deg: 90}}

curve_left: &curve_left
    ~PlacedObject:
        children:
            curve: *curve_right
        spatial_relations: 
            curve: {~SE2Transform: {theta_deg: 270}}
             

1way: &1way
     ~PlacedObject:
        children:
            go_right: *go_right
            go_left: *go_left
            go_straight: *go_straight
            
4way:
    ~PlacedObject:
        children:
            a: *1way
            b: *1way
            c: *1way
            d: *1way
        spatial_relations:
            a: {~SE2Transform: {theta_deg: 0}}
            b: {~SE2Transform: {theta_deg: 90}}
            c: {~SE2Transform: {theta_deg: 180}}
            d: {~SE2Transform: {theta_deg: 270}}
            
#    |    |        
# ---      ----
#  
# -----------

3way_left: &3way_left
    ~PlacedObject:
        children:
            west_go_straight: *go_straight
            west_go_left: *go_left
            north_go_left: *go_left
            north_go_right: *go_right
            east_go_right: *go_right
            east_go_straight: *go_straight
        spatial_relations:
            west_go_straight: {~SE2Transform: {theta_deg: 0}}
            west_go_left: {~SE2Transform: {theta_deg: 0}}
            north_go_left: {~SE2Transform: {theta_deg: -90}}
            north_go_right: {~SE2Transform: {theta_deg: -90}}
            east_go_right: {~SE2Transform: {theta_deg: 180}}
            east_go_straight: {~SE2Transform: {theta_deg: 180}}

3way_right:
    ~PlacedObject:  
        children:
            template: *3way_left           
        spatial_relations:
            template: {~SE2Transform: {theta_deg: 180}}   
            
 

"""


@memoized_reset
def load_tile_types():
    s = yaml.load(data, Loader=yaml.SafeLoader)
    templates = Serializable.from_json_dict(s)
    return templates



from duckietown_world import GenericSequence, SampledSequence
from zuper_commons.logs import ZLogger
from zuper_typing import can_be_used_as2

logger = ZLogger(__name__)


def test_typing():
    s = SampledSequence.from_iterator([(0, "a"), (1, "b"), (2, "c")], str)
    T = GenericSequence[object]

    r = can_be_used_as2(type(s), T)
    logger.info(r=r)
    assert isinstance(s, T)



# -*- coding: utf-8 -*-
import functools


__all__ = ["memoized_reset"]


# def memoize_simple(obj):
# # TODO: make sure it's not iterator
#     cache = obj.cache = {}
#
#     def memoizer(f, *args, **kwargs):
#         key = (args, frozendict2(kwargs))
#         if key not in cache:
#             cache[key] = f(*args, **kwargs)
#             # print('memoize: %s %d storage' % (obj, len(cache)))
#
#         return cache[key]
#
#     return decorator(memoizer, obj)


class memoized_reset(object):
    """Decorator that caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned, and
    not re-evaluated.
    """

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args, **kwargs):
        key = args + tuple(kwargs.items())
        try:
            res = self.cache[key]
            # print('using cache for %s' % self.func)
            return res
        except KeyError:
            value = self.func(*args, **kwargs)
            self.cache[key] = value
            return value
        except TypeError:
            # uncachable -- for instance, passing a list as an argument.
            # Better to not cache than to blow up entirely.
            return self.func(*args, **kwargs)

    def __repr__(self):
        """Return the function's docstring."""
        return self.func.__doc__

    def __get__(self, obj, objtype):  # @UnusedVariable
        """Support instance methods."""
        fn = functools.partial(self.__call__, obj)
        fn.reset = self._reset
        return fn

    def _reset(self):
        self.cache = {}



import math
from typing import Tuple, Any

import numpy as np

from duckietown_world.world_duckietown.types import TSE2v
from .platform_dynamics import PlatformDynamicsFactory, PlatformDynamics

__all__ = ["ApplyDelay"]


class ApplyDelay(PlatformDynamicsFactory):
    factory: PlatformDynamicsFactory

    def __init__(self, factory, delay: float, u0):
        self.factory = factory
        self.delay = delay
        self.u0 = u0

    def initialize(self, c0: TSE2v, t0: float = 0, seed: int = None) -> PlatformDynamics:
        """
            Returns the dynamics initalized at a certain configuration.

            :param c0: configuration in TSE2
            :param t0: time in which to sstart
            :param seed: seed for a possible random number generator

        """
        state0 = self.factory.initialize(c0, t0, seed)
        return DelayedDynamics(state0, self.delay, t0, self.u0)


class DelayedDynamics(PlatformDynamics):
    state: PlatformDynamics

    def __init__(self, state, delay, t0, u0, commands=None, timestamps=None):
        self.state = state
        self.commands = commands or []
        self.timestamps = timestamps or []
        self.t = t0
        self.delay = delay
        self.u0 = u0

        if self.timestamps:
            i, _, _ = self.get_commands_at(self.timestamps[-1] - self.delay)
            self.commands = self.commands[i:]
            self.timestamps = self.timestamps[i:]

    def get_commands_at(self, t) -> Tuple[int, float, Any]:
        if t < self.timestamps[0]:
            return 0, 0, self.u0

        a = np.array(self.timestamps)
        idx = np.searchsorted(a, t, side="left")

        if idx > 0 and (idx == len(a) or math.fabs(t - a[idx - 1]) < math.fabs(t - a[idx])):
            return idx, a[idx], self.commands[idx - 1]
        else:
            return idx, a[idx], self.commands[idx]

    def integrate(self, dt: float, commands) -> "DelayedDynamics":
        """
            Returns the result of applying commands for dt.

            :param dt > 0: time interval
            :param commands: class-specific commands
            :return: the next state
        """
        self.commands.append(commands)
        self.timestamps.append(self.t)
        self.t += dt
        i, told, use_commands = self.get_commands_at(self.t - self.delay)
        # print(f't = {self.t}, t old = {told}'  )
        state2 = self.state.integrate(dt, use_commands)

        next = DelayedDynamics(
            state2,
            self.delay,
            self.t,
            self.u0,
            commands=list(self.commands),
            timestamps=list(self.timestamps),
        )
        return next

    def TSE2_from_state(self) -> TSE2v:
        return self.state.TSE2_from_state()



import os

from duckietown_world import draw_static

d = "out"


def ipython_draw_html(po, outdir=None, area=None):
    from IPython.display import IFrame, display

    if outdir is None:
        outdir = os.path.join(d, "ipython_draw_html", "%s" % id(po))
    draw_static(po, outdir, area=area)

    iframe = IFrame(src=outdir + "/drawing.html", width="100%", height=600)
    # noinspection PyTypeChecker
    display(iframe)
    return iframe


def ipython_draw_svg(m, outdir=None):
    from IPython.display import SVG, display

    if outdir is None:
        outdir = os.path.join(d, "ipython_draw_svg", "%s" % id(m))
    draw_static(m, outdir)

    svg = SVG(os.path.join(outdir, "drawing.svg"))
    # noinspection PyTypeChecker
    display(svg)
    return svg



import os

import yaml
import numpy as np

import geometry as geo
from comptests import comptest, run_module_tests, get_comptests_output_dir
from duckietown_world import (
    PWMCommands,
    SampledSequence,
    draw_static,
    SE2Transform,
    DB18,
    construct_map,
)
from duckietown_world.seqs.tsequence import SampledSequenceBuilder
from duckietown_world.svg_drawing.misc import TimeseriesPlot
from duckietown_world.world_duckietown.pwm_dynamics import get_DB18_nominal
from duckietown_world.world_duckietown.types import TSE2v, se2v
from duckietown_world.world_duckietown.utils import get_velocities_from_sequence


@comptest
def test_pwm1():
    parameters = get_DB18_nominal(delay=0)

    # initial configuration
    init_pose = np.array([0, 0.8])
    init_vel = np.array([0, 0])

    q0 = geo.SE2_from_R2(init_pose)
    v0 = geo.se2_from_linear_angular(init_vel, 0)
    tries = {
        "straight_50": (PWMCommands(+0.5, 0.5)),
        "straight_max": (PWMCommands(+1.0, +1.0)),
        "straight_over_max": (PWMCommands(+1.5, +1.5)),
        "pure_left": (PWMCommands(motor_left=-0.5, motor_right=+0.5)),
        "pure_right": (PWMCommands(motor_left=+0.5, motor_right=-0.5)),
        "slight-forward-left": (PWMCommands(motor_left=0, motor_right=0.25)),
        "faster-forward-right": (PWMCommands(motor_left=0.5, motor_right=0)),
        # 'slight-right': (PWMCommands(-0.1, 0.1)),
    }
    dt = 0.03
    t_max = 10

    map_data_yaml = """

    tiles:
    - [floor/W,floor/W, floor/W, floor/W, floor/W] 
    - [straight/W   , straight/W   , straight/W, straight/W, straight/W]
    - [floor/W,floor/W, floor/W, floor/W, floor/W]
    tile_size: 0.61
    """

    map_data = yaml.load(map_data_yaml)

    root = construct_map(map_data)

    timeseries = {}
    for id_try, commands in tries.items():
        seq = integrate_dynamics(parameters, q0, v0, dt, t_max, commands)

        ground_truth = seq.transform_values(lambda t: SE2Transform.from_SE2(t[0]))
        poses = seq.transform_values(lambda t: t[0])
        velocities = get_velocities_from_sequence(poses)
        linear = velocities.transform_values(linear_from_se2)
        angular = velocities.transform_values(angular_from_se2)
        # print(linear.values)
        # print(angular.values)
        root.set_object(id_try, DB18(), ground_truth=ground_truth)

        sequences = {}
        sequences["motor_left"] = seq.transform_values(lambda _: commands.motor_left)
        sequences["motor_right"] = seq.transform_values(lambda _: commands.motor_right)
        plots = TimeseriesPlot(f"{id_try} - PWM commands", "pwm_commands", sequences)
        timeseries[f"{id_try} - commands"] = plots

        sequences = {}
        sequences["linear_velocity"] = linear
        sequences["angular_velocity"] = angular
        plots = TimeseriesPlot(f"{id_try} - Velocities", "velocities", sequences)
        timeseries[f"{id_try} - velocities"] = plots

    outdir = os.path.join(get_comptests_output_dir(), "together")
    draw_static(root, outdir, timeseries=timeseries)


def linear_from_se2(x: se2v) -> float:
    linear, _ = geo.linear_angular_from_se2(x)
    return linear[0]


def angular_from_se2(x: se2v) -> float:
    _, angular = geo.linear_angular_from_se2(x)
    return angular


def integrate_dynamics(factory, q0, v0, dt, t_max, fixed_commands) -> SampledSequence[TSE2v]:
    # starting time
    t0 = 0
    c0 = q0, v0
    state = factory.initialize(c0=c0, t0=t0)
    ssb = SampledSequenceBuilder[TSE2v]()
    ssb.add(t0, state.TSE2_from_state())
    n = int(t_max / dt)
    for i in range(n):
        state = state.integrate(dt, fixed_commands)
        t = t0 + (i + 1) * dt
        ssb.add(t, state.TSE2_from_state())
    return ssb.as_sequence()


if __name__ == "__main__":
    run_module_tests()



# coding=utf-8
from dataclasses import dataclass
from typing import Iterator, Tuple

import numpy as np
from geometry import extract_pieces
from svgwrite.container import Use

from duckietown_world import logger
from duckietown_world.geo import (
    Matrix2D,
    PlacedObject,
    RectangularArea,
    SE2Transform,
    Transform,
    TransformSequence,
)
from duckietown_world.geo.measurements_utils import (
    iterate_by_class,
    IterateByTestResult,
)
from duckietown_world.seqs import SampledSequence
from duckietown_world.svg_drawing import data_encoded_for_src, draw_axes, draw_children
from duckietown_world.svg_drawing.misc import mime_from_fn
from .lane_segment import LanePose, LaneSegment
from .tile_coords import TileCoords
from .types import SE2v

__all__ = ["Tile", "GetLanePoseResult", "get_lane_poses", "create_lane_highlight"]

from .utils import relative_pose


class SignSlot(PlacedObject):
    """ Represents a slot where you can put a sign. """

    L = 0.065 / 0.61

    def get_footprint(self):
        L = SignSlot.L
        return RectangularArea([-L / 2, -L / 2], [L / 2, L / 2])

    def draw_svg(self, drawing, g):
        L = SignSlot.L

        rect = drawing.rect(
            insert=(-L / 2, -L / 2),
            size=(L, L),
            fill="none",
            # style='opacity:0.4',
            stroke_width="0.005",
            stroke="pink",
        )
        g.add(rect)
        draw_axes(drawing, g, 0.04)


def get_tile_slots():
    LM = 0.5  # half tile
    # tile_offset
    to = 0.20
    # tile_curb
    tc = 0.05

    positions = {
        0: (+to, +tc),
        1: (+tc, +to),
        2: (-tc, +to),
        3: (-to, +tc),
        4: (-to, -tc),
        5: (-tc, -to),
        6: (+tc, -to),
        7: (+to, -tc),
    }

    po = PlacedObject()
    for i, (x, y) in positions.items():
        name = str(i)
        # if name in self.children:
        #     continue

        sl = SignSlot()
        # theta = np.deg2rad(theta_deg)
        theta = 0
        t = SE2Transform((-LM + x, -LM + y), theta)
        po.set_object(name, sl, ground_truth=t)
    return po


class Tile(PlacedObject):
    def __init__(self, kind, drivable, **kwargs):
        # noinspection PyArgumentList
        PlacedObject.__init__(self, **kwargs)
        self.kind = kind
        self.drivable = drivable

        from duckietown_world.world_duckietown.map_loading import get_texture_file

        try:
            self.fn = get_texture_file(kind)
        except KeyError:
            msg = "Cannot find texture for %s" % kind
            logger.warning(msg)

            self.fn = None
        # if kind in ['asphalt']:
        if not "slots" in self.children:
            slots = get_tile_slots()
            self.set_object("slots", slots, ground_truth=SE2Transform.identity())

    def _copy(self):

        return type(self)(
            self.kind,
            self.drivable,
            children=dict(self.children),
            spatial_relations=dict(self.spatial_relations),
        )

    def params_to_json_dict(self):
        return dict(kind=self.kind, drivable=self.drivable)

    def get_footprint(self):
        return RectangularArea([-0.5, -0.5], [0.5, 0.5])

    def draw_svg(self, drawing, g):
        T = 0.562 / 0.585
        S = 1.0
        rect = drawing.rect(insert=(-S / 2, -S / 2), size=(S, S), fill="#222", stroke="none")
        g.add(rect)

        if self.fn:
            # texture = get_jpeg_bytes(self.fn)
            with open(self.fn, "rb") as _:
                texture = _.read()
            if b"git-lfs" in texture:
                msg = f"The file {self.fn} is a Git LFS pointer. Repo not checked out correctly."
                raise Exception(msg)
            # print(f'drawing defs {drawing.defs}')

            ID = f"texture-{self.kind}"

            for img in drawing.defs.elements:
                if img.attribs.get("id", None) == ID:
                    break
            else:

                href = data_encoded_for_src(texture, mime_from_fn(self.fn))
                img = drawing.image(
                    href=href,
                    size=(T, T),
                    insert=(-T / 2, -T / 2),
                    style="transform: rotate(90deg) scaleX(-1)  rotate(-90deg) ",
                )
                img.attribs["class"] = "tile-textures"

                img.attribs["id"] = ID
                drawing.defs.add(img)

            use = Use(f"#{ID}")
            g.add(use)
        #
        # if draw_directions_lanes:
        #     if self.kind != 'floor':
        #         start = (-0.5, -0.25)
        #         end = (+0, -0.25)
        #         line = drawing.line(start=start, end=end, stroke='blue', stroke_width='0.01')
        #         g.add(line)

        draw_axes(drawing, g)

        draw_children(drawing, self, g)


@dataclass
class GetLanePoseResult:
    tile: Tile
    tile_fqn: Tuple[str, ...]
    tile_transform: TransformSequence
    tile_relative_pose: Matrix2D
    lane_segment: LaneSegment
    lane_segment_fqn: Tuple[str, ...]
    lane_pose: LanePose
    lane_segment_relative_pose: Matrix2D
    tile_coords: TileCoords
    lane_segment_transform: TransformSequence
    center_point: Matrix2D


def get_lane_poses(dw: PlacedObject, q: SE2v, tol=0.000001) -> Iterator[GetLanePoseResult]:
    for it in iterate_by_class(dw, Tile):
        assert isinstance(it, IterateByTestResult), it
        assert isinstance(it.object, Tile), it.object
        tile = it.object
        tile_fqn = it.fqn
        tile_transform = it.transform_sequence
        for _ in tile_transform.transforms:
            if isinstance(_, TileCoords):
                tile_coords = _
                break
        else:
            msg = "Could not find tile coords in %s" % tile_transform
            assert False, msg
        # print('tile_transform: %s' % tile_transform.asmatrix2d().m)
        tile_relative_pose = relative_pose(tile_transform.asmatrix2d().m, q)
        p = translation_from_O3(tile_relative_pose)
        # print('tile_relative_pose: %s' % tile_relative_pose)
        if not tile.get_footprint().contains(p):
            continue
        nresults = 0
        for it2 in iterate_by_class(tile, LaneSegment):
            lane_segment = it2.object
            lane_segment_fqn = tile_fqn + it2.fqn
            assert isinstance(lane_segment, LaneSegment), lane_segment
            lane_segment_wrt_tile = it2.transform_sequence.asmatrix2d()
            lane_segment_relative_pose = relative_pose(lane_segment_wrt_tile.m, tile_relative_pose)
            lane_segment_transform = TransformSequence(
                tile_transform.transforms + it2.transform_sequence.transforms
            )
            lane_pose = lane_segment.lane_pose_from_SE2(lane_segment_relative_pose, tol=tol)

            M = lane_segment_transform.asmatrix2d().m
            center_point = lane_pose.center_point.as_SE2()

            center_point_abs = np.dot(M, center_point)
            center_point_abs_t = Matrix2D(center_point_abs)

            if lane_pose.along_inside and lane_pose.inside and lane_pose.correct_direction:
                yield GetLanePoseResult(
                    tile=tile,
                    tile_fqn=tile_fqn,
                    tile_transform=tile_transform,
                    tile_relative_pose=Matrix2D(tile_relative_pose),
                    lane_segment=lane_segment,
                    lane_segment_relative_pose=Matrix2D(lane_segment_relative_pose),
                    lane_pose=lane_pose,
                    lane_segment_fqn=lane_segment_fqn,
                    lane_segment_transform=lane_segment_transform,
                    tile_coords=tile_coords,
                    center_point=center_point_abs_t,
                )
                nresults += 1

        # if nresults == 0:
        #     msg = 'Could not find any lane in tile %s' % tile_transform
        #     msg += '\ntile_relative_pose: %s' % tile_relative_pose
        #     for it2 in iterate_by_class(tile, LaneSegment):
        #         lane_segment = it2.object
        #         lane_segment_wrt_tile = it2.transform_sequence.as_SE2()
        #         lane_segment_relative_pose = relative_pose(lane_segment_wrt_tile, tile_relative_pose)
        #         lane_pose = lane_segment.lane_pose_from_SE2(lane_segment_relative_pose, tol=tol)
        #
        #         msg += '\n lane_relative: %s' % lane_segment_relative_pose
        #         msg += '\n lane pose: %s' % lane_pose
        #     logger.warning(msg)


def translation_from_O3(pose) -> np.ndarray:
    _, t, _, _ = extract_pieces(pose)
    return t


class GetClosestLane:
    def __init__(self, dw):
        self.no_matches_for = []
        self.dw = dw

    def __call__(self, transform):
        if isinstance(transform, SE2Transform):
            transform = transform.as_SE2()
        poses = list(get_lane_poses(self.dw, transform))
        # if not poses:
        #     self.no_matches_for.append(transform)
        #     return None

        s = sorted(poses, key=lambda _: np.abs(_.lane_pose.relative_heading))
        res = {}
        for i, _ in enumerate(s):
            res[i] = _

        return res


class Anchor(PlacedObject):
    def _copy(self):
        return self._simplecopy()

    def draw_svg(self, drawing, g):
        draw_axes(drawing, g, klass="anchor-axes")
        c = drawing.circle(center=(0, 0), r=0.03, fill="blue", stroke="black", stroke_width=0.001)
        g.add(c)


def create_lane_highlight(poses_sequence: SampledSequence, dw):
    def mapi(v):
        if isinstance(v, SE2Transform):
            return v.as_SE2()
        else:
            return v

    poses_sequence = poses_sequence.transform_values(mapi, np.ndarray)

    lane_pose_results = poses_sequence.transform_values(GetClosestLane(dw), object)

    visualization = PlacedObject()
    dw.set_object("visualization", visualization, ground_truth=SE2Transform.identity())
    for i, (timestamp, name2pose) in enumerate(lane_pose_results):
        for name, lane_pose_result in name2pose.items():
            assert isinstance(lane_pose_result, GetLanePoseResult)
            lane_segment = lane_pose_result.lane_segment
            rt = lane_pose_result.lane_segment_transform
            s = SampledSequence[Transform]([timestamp], [rt])
            visualization.set_object("ls%s-%s-lane" % (i, name), lane_segment, ground_truth=s)
            p = SampledSequence[Transform]([timestamp], [lane_pose_result.center_point])
            visualization.set_object("ls%s-%s-anchor" % (i, name), Anchor(), ground_truth=p)

    return lane_pose_results



# coding=utf-8
import networkx as nx
import numpy as np
from dataclasses import dataclass
from duckietown_world.seqs import GenericSequence
from networkx import MultiDiGraph
from typing import Callable, Iterator, List, Tuple

from .placed_object import FQN, PlacedObject, SpatialRelation
from .rectangular_area import RectangularArea
from .transforms import TransformSequence, VariableTransformSequence

__all__ = [
    "iterate_measurements_relations",
    "get_meausurements_graph",
    "get_extent_points",
    "get_static_and_dynamic",
    "iterate_by_class",
    "IterateByTestResult",
]


def iterate_measurements_relations(po_name: FQN, po: PlacedObject) -> Iterator[Tuple[FQN, SpatialRelation]]:
    assert isinstance(po_name, tuple)
    for sr_name, sr in po.spatial_relations.items():
        a = po_name + sr.a
        b = po_name + sr.b
        klass = type(sr)
        s = klass(a=a, b=b, transform=sr.transform)
        yield po_name + (sr_name,), s

    for child_name, child in po.children.items():
        cname = po_name + (child_name,)
        yield from iterate_measurements_relations(cname, child)


def get_meausurements_graph(po: PlacedObject) -> MultiDiGraph:
    G = MultiDiGraph()
    for name, sr in iterate_measurements_relations((), po):
        a = sr.a
        b = sr.b
        attr_dict = dict(sr=sr)
        G.add_edge(a, b, attr_dict=attr_dict)
    return G


# FlattenResult = namedtuple('FlattenResult', 'po old2new')


def get_static_and_dynamic(po: PlacedObject) -> Tuple[List, List]:
    assert isinstance(po, PlacedObject)

    G = get_flattened_measurement_graph(po)

    static = []

    static.append(())
    dynamic = []
    root_name = ()
    for name in G.nodes():
        if name == root_name:
            continue
        edge_data = G.get_edge_data(root_name, name)

        transform = edge_data["transform_sequence"]
        from duckietown_world.world_duckietown.transformations import is_static

        it_is = is_static(transform)

        if it_is:
            static.append(name)
        else:
            dynamic.append(name)

    return static, dynamic


def get_flattened_measurement_graph(po: PlacedObject, include_root_to_self: bool = False) -> nx.DiGraph:
    from duckietown_world import TransformSequence
    from duckietown_world import SE2Transform

    G = get_meausurements_graph(po)
    G2 = nx.DiGraph()
    root_name = ()
    for name in G.nodes():
        if name == root_name:
            continue
        path = nx.shortest_path(G, (), name)
        transforms = []
        for i in range(len(path) - 1):
            a = path[i]
            b = path[i + 1]
            edges = G.get_edge_data(a, b)

            k = list(edges)[0]
            v = edges[k]
            sr = v["attr_dict"]["sr"].transform

            transforms.append(sr)

        if any(isinstance(_, GenericSequence) for _ in transforms):
            res = VariableTransformSequence(transforms)
        else:
            res = TransformSequence(transforms)
        G2.add_edge(root_name, name, transform_sequence=res)

    if include_root_to_self:
        transform_sequence = SE2Transform.identity()
        G2.add_edge(root_name, root_name, transform_sequence=transform_sequence)

    return G2


@dataclass
class IterateByTestResult:
    fqn: Tuple[str]
    transform_sequence: TransformSequence
    object: PlacedObject
    parents: Tuple[PlacedObject, ...]


# IterateByTestResult = namedtuple('IterateByTestResult', 'fqn transform_sequence object')


def iterate_by_class(po: PlacedObject, klass: type) -> Iterator[IterateByTestResult]:
    t = lambda _: isinstance(_, klass)
    yield from iterate_by_test(po, t)


def iterate_by_test(po: PlacedObject, testf: Callable[[PlacedObject], bool]) -> Iterator[IterateByTestResult]:
    G = get_flattened_measurement_graph(po, include_root_to_self=True)
    root_name = ()

    for name in G:
        ob = po.get_object_from_fqn(name)

        if testf(ob):
            transform_sequence = G.get_edge_data(root_name, name)["transform_sequence"]

            parents = get_parents(po, name)

            yield IterateByTestResult(
                fqn=name, transform_sequence=transform_sequence, object=ob, parents=parents,
            )


def get_parents(root: PlacedObject, child_fqn: FQN) -> Tuple[PlacedObject]:
    parents = []
    n = len(child_fqn)
    for i in range(0, n - 1):
        name_parent: FQN = child_fqn[:i]
        parent = root.get_object_from_fqn(name_parent)
        parents.append(parent)
    return tuple(parents)


def get_extent_points(root: PlacedObject) -> RectangularArea:
    G = get_flattened_measurement_graph(root, include_root_to_self=True)
    points = []

    root_name = ()
    for name in G.nodes():

        transform_sequence = G.get_edge_data(root_name, name)["transform_sequence"]
        extent_points = root.get_object_from_fqn(name).extent_points()
        m2d = transform_sequence.asmatrix2d()
        for _ in extent_points:
            p = np.dot(m2d.m, [_[0], _[1], 1])[:2]
            points.append(p)

    if not points:
        msg = "No points"
        raise ValueError(msg)

    pmin = points[0]
    pmax = points[0]

    for p in points:
        pmin = np.minimum(pmin, p)
        pmax = np.maximum(pmax, p)

    return RectangularArea(pmin, pmax)



from networkx import connected_components

import duckietown_world as dw
import geometry as geo
from comptests import comptest
from duckietown_world.world_duckietown.segmentify import MeetingPoint
from duckietown_world.world_duckietown.tile_utils import get_tile_at_point
from zuper_typing import debug_print


@comptest
def get_lane_at_point_test():
    m: dw.DuckietownMap = dw.load_map("robotarium2")

    p = geo.SE2_from_translation_angle([1.3, 0.3], 0.2)
    r = get_tile_at_point(m, p)
    assert r.i, r.j == (2, 0)


@comptest
def get_lane_at_point_test2():
    m: dw.DuckietownMap = dw.load_map("robotarium2")

    sk2 = dw.get_skeleton_graph(m)
    G0 = sk2.G0
    c = list(connected_components(G0.to_undirected()))
    assert len(c) == 6

    p = list(G0.nodes)[0]
    mp: MeetingPoint = G0.nodes[p]["meeting_point"]
    print(debug_print(mp))


if __name__ == "__main__":
    get_lane_at_point_test2()



# coding=utf-8
__version__ = "6.0.3"

import logging

logging.basicConfig()
logger = logging.getLogger("dt-world")
logger.setLevel(logging.DEBUG)

logger.info("duckietown-world %s" % __version__)

# remove noisy logging
from duckietown_serialization_ds1 import logger as dslogger

dslogger.setLevel(logging.CRITICAL)

from contracts import disable_all

disable_all()

from .geo import *
from .seqs import *
from .svg_drawing import *
from .world_duckietown import *



# coding=utf-8
import geometry
from abc import ABCMeta, abstractmethod
from typing import NewType

import numpy as np
from duckietown_serialization_ds1 import Serializable

import geometry as geo
from contracts import contract, new_contract
from duckietown_world.seqs.tsequence import Timestamp
from duckietown_world.seqs import GenericSequence


__all__ = ["TransformSequence", "Transform", "SE2Transform", "Scale2D", "Matrix2D"]

SE2value = NewType("SE2value", np.ndarray)


class Transform(metaclass=ABCMeta):
    @abstractmethod
    def asmatrix2d(self) -> "Matrix2D":
        """ """

    def as_SE2(self):
        m = self.asmatrix2d().m
        try:
            geo.SE2.belongs(m)
        except:
            msg = "Cannot convert %s" % type(self).__name__
            msg += "m = %s" % m
            raise Exception(msg)
        return m


new_contract("Transform", Transform)


class TransformSequence(Transform):

    # @contract(transforms='list[>=1](Transform)')
    def __init__(self, transforms):
        self.transforms = list(transforms)

    def asmatrix2d(self):
        ms = [_.asmatrix2d() for _ in self.transforms]
        # print('ms: %s' % ms)
        result = ms[0].m

        for mi in ms[1:]:
            result = np.dot(result, mi.m)
        return Matrix2D(result)

    def __repr__(self):
        return "TransformSequence(%s)" % self.transforms


class VariableTransformSequence(TransformSequence, GenericSequence[Transform]):  # XXX
    def at(self, t: Timestamp):
        res = []
        for transform in self.transforms:
            if isinstance(transform, GenericSequence):
                transform = transform.at(t)
            res.append(transform)
        return TransformSequence(res)

    def get_end(self):
        return None

    def get_start(self):
        return None

    def get_sampling_points(self):
        pass


class SE2Transform(Transform, Serializable):
    @contract(p="seq[2](float|int)")
    def __init__(self, p, theta):
        self.p = np.array(p, dtype="float64")
        self.theta = float(theta)

    def __repr__(self):
        d = np.rad2deg(self.theta)
        return "SE2Transform(%s,%.1f)" % (self.p.tolist(), d)

    @classmethod
    def identity(cls) -> "SE2Transform":
        return SE2Transform([0.0, 0.0], 0.0)

    @classmethod
    def from_SE2(cls, q: SE2value) -> "SE2Transform":
        """ From a matrix """
        translation, angle = geo.translation_angle_from_SE2(q)
        return SE2Transform(translation, angle)

    def params_to_json_dict(self):
        res = {}
        if np.allclose(self.theta, 0):
            pass
        else:
            for a in [-270, -180, -90, -45, +45, 90, 180, 270]:
                if np.allclose(a, np.rad2deg(self.theta)):
                    res["theta_deg"] = a
                    break
            else:
                res["theta"] = self.theta
        if np.allclose(np.linalg.norm(self.p), 0):
            pass
        else:
            res["p"] = self.p
        return res

    @classmethod
    def params_from_json_dict(cls, d):
        if d is None:
            d = {}
        p = d.pop("p", [0.0, 0.0])

        if "theta" in d:
            theta = d.pop("theta")

        elif "theta_deg" in d:
            theta_deg = d.pop("theta_deg")
            theta = np.deg2rad(theta_deg)
        else:
            theta = 0.0

        return dict(p=p, theta=theta)

    def as_SE2(self) -> SE2value:

        M = geometry.SE2_from_translation_angle(self.p, self.theta)
        return M

    def asmatrix2d(self):
        M = self.as_SE2()
        return Matrix2D(M)


new_contract("SE2Transform", SE2Transform)


class Scale2D(Transform, Serializable):
    def __init__(self, scale):
        self.scale = scale

    def asmatrix2d(self):
        S = self.scale
        M = np.array([[S, 0, 0], [0, S, 0], [0, 0, 1]])
        return Matrix2D(M)


class Matrix2D(Transform, Serializable):
    def __init__(self, m):
        self.m = np.array(m, "float32")
        assert self.m.shape == (3, 3)

    def asmatrix2d(self):
        return self



import os
import argparse
from pathlib import Path
import os
import yaml
import pprint


class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def error(string: str):
    print(
        bcolors.FAIL
        + "--------------------------------------------------------------------------------------"
        + bcolors.ENDC
    )
    print("\n\n" + bcolors.FAIL + string + bcolors.ENDC + "\n\n")

    print(
        bcolors.FAIL
        + "--------------------------------------------------------------------------------------"
        + bcolors.ENDC
    )


def warning(string: str):
    print("\n\n" + bcolors.WARNING + string + bcolors.ENDC + "\n\n")


def header(string: str):
    print(bcolors.OKGREEN + string + bcolors.ENDC)


def separater(string: str = None):
    if string == None:
        warning("--------------------------------------------------------------------------------------")
    else:
        warning(
            "--------------------------------------------------------------------------------------\n%s"
            % string
        )


def write_yaml(fn, data):
    with open(fn, "w") as f:
        yaml.dump(data, f)


def load_yaml_file(fn: str):
    with open(fn) as f:
        data = f.read()
    return yaml.load(data, Loader=yaml.SafeLoader)


def create_empty_map_file():
    empty_map = {"objects": {}, "tiles": {}, "tile_size": 0.585, "version": 2}
    return empty_map


def input_int(string):
    int_input = input(string)
    try:
        int_input = int(int_input)
    except Exception as e:
        error("Error : %s" % e)
        return None
    return int_input


def input_float(string):
    float_input = input(string)
    try:
        float_input = float(float_input)
    except Exception as e:
        error("Error : %s" % e)
        return None
    return float_input


class Apriltag:
    def __init__(self, tag_id, x, y, angle):
        self.x = x
        self.tag_id = tag_id
        self.y = y
        self.angle = angle

    def to_dict(self):
        return {
            "kind": "floor_tag",
            "pose": {"~SE2Transform": {"p": [self.x, self.y], "theta_deg": self.angle}},
            "tag": {"~TagInstance": {"family": "36h11", "size": 0.08, "tag_id": self.tag_id}},
        }


class Apriltag_measurer:
    def __init__(self, full_path, new_file, modify_file):
        self.name = "Mouahaha"
        if new_file and modify_file:
            error("Can not at the same time create and modify a file")
            exit()
        elif modify_file:
            self.map_yaml = load_yaml_file(full_path)
            if "objects" not in self.map_yaml:
                self.map_yaml["objects"] = dict()
            self.ground_tag_dict = self.get_ground_tag_dict()
        else:
            self.map_yaml = create_empty_map_file()
            self.ground_tag_dict = dict()

        self.x_offset = 0.012
        self.y_offset = 0.012
        self.tile_size = float(self.map_yaml["tile_size"])

    def get_ground_tag_dict(self):
        apriltag_dict = dict()
        for name, at_dict in self.map_yaml["objects"].items():
            if at_dict["kind"] == "floor_tag":
                tag_id = int(at_dict["tag"]["~TagInstance"]["tag_id"])
                apriltag_dict[tag_id] = name

        return apriltag_dict

    def get_at_dict(self, at_number):
        x_count = input_int("give x coordinate of tile (first tile at origin is (0,0)) : ")
        if x_count is not None:
            y_count = input_int("give y coordinate of tile (first tile at origin is (0,0)) : ")
            if y_count is not None:
                x_measure = input_float("give x measure in meters from interior side of tile : ")
                if x_measure is not None:
                    y_measure = input_float("give y measure in meters from interior side of tile : ")
                    if y_measure is not None:
                        angle_measure = input_int(
                            "give angle of apriltag in degrees (turned as readable from origin is 0) : "
                        )
                        if angle_measure is not None:
                            complete_x = self.tile_size * x_count + x_measure + self.x_offset
                            complete_y = self.tile_size * y_count + y_measure + self.y_offset
                            apriltag = Apriltag(at_number, complete_x, complete_y, angle_measure)
                            return apriltag.to_dict()
        return {}

    def update_apriltags(self):
        result_map_yaml = self.map_yaml
        while True:
            header("Measure a new april tag. Enter -1 to exit ")
            at_number = input("April tag id : ")
            try:
                at_number = int(at_number)
            except Exception as e:
                error("Error : %s " % e)
                continue
            if at_number < 0:
                break
            if at_number < 300 or at_number >= 400:
                warning("This april tag number (%i) is not in the 300-399 range" % at_number)
                continue_anyway = input("Are you sure you want to proceed? (y/n) : ")
                if continue_anyway != "y":
                    continue

            if at_number in self.ground_tag_dict:
                print("The april tag %i is already present" % at_number)
                modify = input("Do you want to modify its coordinates ? (y/n) : ")
                if modify == "y":
                    print("modifying april tag %i" % at_number)
                    at_dict = self.get_at_dict(at_number)
                    new_name = "tag%i" % at_number
                    result_map_yaml["objects"].pop(self.ground_tag_dict[at_number])
                    result_map_yaml["objects"][new_name] = at_dict
                    self.ground_tag_dict[at_number] = new_name

                else:
                    print("not modifying april tag %i" % at_number)
                    continue
            else:
                print("Adding new april tag %i" % at_number)
                at_dict = self.get_at_dict(at_number)
                new_name = "tag%i" % at_number
                result_map_yaml["objects"][new_name] = at_dict
                self.ground_tag_dict[at_number] = new_name
        return result_map_yaml


MAP_PATH = "src/duckietown_world/data/gd1/maps"


def main():
    parser = argparse.ArgumentParser(description="")

    parser.add_argument("map_name", type=str, help="name of the map to create or change")

    args = parser.parse_args()

    map_name = args.map_name

    if not map_name.lower().endswith((".yaml", ".yml")):
        map_name = map_name.split(".")[0] + ".yaml"

    header("--------------\n")
    header("WELCOME FRIEND")
    header("\n--------------")

    abs_path = os.path.abspath(MAP_PATH)

    full_path = os.path.join(abs_path, map_name)

    new_file = False
    modify_file = False

    if not os.path.isfile(full_path):
        if not os.path.isdir(abs_path):
            error("the directory %s does not exist" % abs_path)
        else:
            warning("file %s does not exist,\nbut directory %s is there." % (full_path, abs_path))
            new_file_asked = input("Do you want to create file %s ? (y/n) : " % map_name)
            if new_file_asked == "y":
                new_file = False
            else:
                error("Program is closing as map file doesnt exist and is not to be created")
                exit()
    else:
        header("File %s is already present." % map_name)
        continue_file = input(
            "Would you like to load and modify april tag positions for %s ? (y/n) : " % map_name
        )
        if continue_file == "y":
            modify_file = True
        else:
            error("Program is closing as map file exists and is not to be modified")
            exit()

    apriltag_measurer = Apriltag_measurer(full_path, new_file, modify_file)

    result_map_yaml = apriltag_measurer.update_apriltags()
    pprint.pprint(result_map_yaml["objects"])

    separater()
    header("Inputing of april tag position done")
    separater()
    while True:
        overwrite = input("Are you sure you want to overwrite %s ? (y/n) : " % map_name)
        if overwrite == "y":
            write_yaml(full_path, result_map_yaml)
            separater()
            header("File %s saved at %s. Exiting now." % (map_name, full_path))
            separater()
            exit()
        elif overwrite == "n":
            error("Did not overwrite %s. Exiting now." % map_name)
            exit()
        else:
            warning("Please enter 'y' or 'n'")


if __name__ == "__main__":
    main()



# coding=utf-8
from typing import List

import numpy as np
import svgwrite

import geometry as geo
from contracts import contract, new_contract
from duckietown_serialization_ds1 import Serializable
from duckietown_serialization_ds1.serialization1 import as_json_dict
from duckietown_world.geo import PlacedObject, SE2Transform
from duckietown_world.geo.transforms import SE2value
from duckietown_world.utils import memoized_reset, SE2_apply_R2, SE2_interpolate
from zuper_commons.types import check_isinstance
from .utils import relative_pose

__all__ = ["LaneSegment", "LanePose"]


class LanePose(Serializable):
    """ Very detailed information about the "position in the lane". """

    ### Am I inside the lane? If not why not?

    # am I "inside" the lane?
    inside: bool
    # am I "inside" considering the lateral position?
    lateral_inside: bool
    # if not, am I outside on the left or right?
    outside_left: bool
    outside_right: bool
    # am I "inside" considering the longitudinal position?
    along_inside: bool
    # if not, am I outside before? (along_lane < 0)
    along_before: bool  #
    # or am I outside after? (along_lane > L)
    along_after: bool

    ### The three main coordinates.
    # Lateral, where 0 = lane center, positive to the left
    lateral: float
    # Longitudinal, along the lane. Starts at 0, positive going forward
    along_lane: float
    # Heading direction: 0 means aligned with the direction of the lane
    relative_heading: float

    # am I going in the right direction?
    correct_direction: bool

    ## Some useful pre-computed quantities

    # lateral position of closest left lane boundary
    lateral_left: float
    # lateral position of closest right lane boundary
    lateral_right: float

    # The distance from us to the left lane boundary
    distance_from_left: float
    # The distance from us to the right lane boundary
    distance_from_right: float
    # The distance from us to the center = abs(lateral)
    distance_from_center: float

    # center_point: anchor point on the center of the lane
    center_point: SE2Transform

    def __init__(
        self,
        inside,
        outside_left,
        outside_right,
        along_lane,
        lateral,
        relative_heading,
        distance_from_left,
        distance_from_right,
        along_inside,
        along_before,
        along_after,
        lateral_left,
        lateral_right,
        lateral_inside,
        distance_from_center,
        center_point,
        correct_direction,
    ):
        self.inside = inside
        self.lateral_inside = lateral_inside
        self.outside_left = outside_left
        self.outside_right = outside_right
        self.lateral = lateral
        self.lateral_left = lateral_left
        self.lateral_right = lateral_right

        self.along_inside = along_inside
        self.along_before = along_before
        self.along_after = along_after
        self.along_lane = along_lane

        self.relative_heading = relative_heading

        self.distance_from_left = distance_from_left
        self.distance_from_right = distance_from_right
        self.distance_from_center = distance_from_center
        self.center_point = center_point
        self.correct_direction = correct_direction


def almost_equal(a, b):
    return (np.abs(a - b)) < 1e-7


class LaneSegment(PlacedObject):
    def __init__(self, width: float, control_points: List[SE2Transform], *args, **kwargs):
        # noinspection PyArgumentList
        PlacedObject.__init__(self, *args, **kwargs)
        self.width = float(width)
        assert len(control_points) >= 2, control_points
        self.control_points = control_points

        for p in control_points:
            check_isinstance(p, SE2Transform)

        for i in range(len(control_points) - 1):
            a = control_points[i]
            b = control_points[i + 1]
            ta, _ = geo.translation_angle_from_SE2(a.as_SE2())
            tb, _ = geo.translation_angle_from_SE2(b.as_SE2())
            d = np.linalg.norm(ta - tb)
            if d < 0.001:
                msg = "Two points are two close: \n%s\n%s" % (a, b)
                raise ValueError(msg)

    def _copy(self):
        return self._simplecopy(width=self.width, control_points=self.control_points)

    @memoized_reset
    def get_lane_lengths(self):
        res = []
        for i in range(len(self.control_points) - 1):
            p0 = self.control_points[i]
            p1 = self.control_points[i + 1]
            sd = get_distance_two(p0.as_SE2(), p1.as_SE2())
            res.append(sd)
        return res

    def get_lane_length(self):
        return sum(self.get_lane_lengths())

    def lane_pose_random(self):
        along_lane = np.random.uniform(0, self.get_lane_length())
        W = self.width
        lateral = np.random.uniform(-W / 2, +W / 2)
        angle = np.random.uniform(-np.pi / 2, +np.pi / 2)
        return self.lane_pose(along_lane=along_lane, lateral=lateral, relative_heading=angle)

    def lane_pose(self, along_lane: float, lateral: float, relative_heading: float) -> LanePose:
        beta = self.beta_from_along_lane(along_lane)
        center_point = self.center_point(beta)

        W2 = self.width / 2
        lateral_inside = -W2 <= lateral <= W2
        outside_right = lateral < -W2
        outside_left = W2 < lateral
        distance_from_left = np.abs(+W2 - lateral)
        distance_from_right = np.abs(-W2 - lateral)
        distance_from_center = np.abs(lateral)

        L = self.get_lane_length()
        along_inside = 0 <= along_lane < L
        along_before = along_lane < 0
        along_after = along_lane > L
        inside = lateral_inside and along_inside

        correct_direction = np.abs(relative_heading) <= np.pi / 2
        return LanePose(
            inside=inside,
            lateral_inside=lateral_inside,
            outside_left=outside_left,
            outside_right=outside_right,
            distance_from_left=distance_from_left,
            distance_from_right=distance_from_right,
            relative_heading=relative_heading,
            along_inside=along_inside,
            along_before=along_before,
            along_after=along_after,
            along_lane=along_lane,
            lateral=lateral,
            lateral_left=W2,
            lateral_right=-W2,
            distance_from_center=distance_from_center,
            center_point=SE2Transform.from_SE2(center_point),
            correct_direction=correct_direction,
        )

    def params_to_json_dict(self):
        return dict(width=self.width, control_points=as_json_dict(self.control_points))

    def extent_points(self):
        return self.lane_profile()

    def lane_pose_from_SE2Transform(self, qt: SE2Transform, tol=0.001) -> LanePose:
        return self.lane_pose_from_SE2(qt.as_SE2(), tol=tol)

    def lane_pose_from_SE2(self, q, tol=0.001) -> LanePose:
        return (
            self.is_straight()
            and self.lane_pose_from_SE2_straight(q)
            or self.lane_pose_from_SE2_generic(q, tol=tol)
        )

    # return condition ? if_true : if_false
    # return condition and if_true or if_false

    def is_straight(self):
        if not len(self.control_points) == 2:
            return False

        cp1 = self.control_points[0]
        cp2 = self.control_points[1]

        r = relative_pose(cp1.as_SE2(), cp2.as_SE2())
        p, theta = geo.translation_angle_from_SE2(r)
        return almost_equal(theta, 0) and almost_equal(p[1], 0)

    @contract(q="euclidean2")
    def lane_pose_from_SE2_straight(self, q):
        cp1 = self.control_points[0]
        rel = relative_pose(cp1.as_SE2(), q)
        tas = geo.translation_angle_scale_from_E2(rel)
        lateral = tas.translation[1]
        along_lane = tas.translation[0]
        relative_heading = tas.angle
        return self.lane_pose(relative_heading=relative_heading, lateral=lateral, along_lane=along_lane)

    @contract(q="euclidean2")
    def lane_pose_from_SE2_generic(self, q, tol=0.001):
        p, _, _ = geo.translation_angle_scale_from_E2(q)

        beta, q0 = self.find_along_lane_closest_point(p, tol=tol)
        along_lane = self.along_lane_from_beta(beta)
        rel = relative_pose(q0, q)

        r, relative_heading, _ = geo.translation_angle_scale_from_E2(rel)
        lateral = r[1]
        # extra = r[0]
        # along_lane -= extra
        # logger.info('ms: %s' % r[0])
        return self.lane_pose(relative_heading=relative_heading, lateral=lateral, along_lane=along_lane)

    def find_along_lane_closest_point(self, p, tol=0.001):
        def get_delta(beta):
            q0 = self.center_point(beta)
            t0, _ = geo.translation_angle_from_SE2(q0)
            d = np.linalg.norm(p - t0)

            d1 = np.array([0, -d])
            p1 = SE2_apply_R2(q0, d1)

            d2 = np.array([0, +d])
            p2 = SE2_apply_R2(q0, d2)

            D2 = np.linalg.norm(p2 - p)
            D1 = np.linalg.norm(p1 - p)
            res = np.minimum(D1, D2)
            # print('%10f: q %s %f' % (beta, geo.SE2.friendly(q0), res))
            return res

        import scipy.optimize

        bracket = (-1.0, len(self.control_points))
        res = scipy.optimize.minimize_scalar(get_delta, bracket=bracket, tol=tol)
        # print(res)
        beta = res.x
        q0 = self.center_point(beta)
        return beta, q0

    def SE2Transform_from_lane_pose(self, lane_pose: LanePose) -> SE2Transform:
        beta = self.beta_from_along_lane(lane_pose.along_lane)
        # logger.info('beta: %s' % beta)
        center_point = self.center_point(beta)
        delta = np.array([0, lane_pose.lateral])

        rel = geo.SE2_from_translation_angle(delta, lane_pose.relative_heading)
        # logger.info('rel: %s' % geo.SE2.friendly(rel))
        res = geo.SE2.multiply(center_point, rel)

        return SE2Transform.from_SE2(res)

    def along_lane_from_beta(self, beta):
        lengths = self.get_lane_lengths()
        if beta < 0:
            return beta
        elif beta >= len(self.control_points) - 1:
            rest = beta - (len(self.control_points) - 1)
            return sum(lengths) + rest
        else:
            i = int(np.floor(beta))

            rest = beta - i

            res = sum(lengths[:i]) + lengths[i] * rest

            return res

    def beta_from_along_lane(self, along_lane):
        lengths = self.get_lane_lengths()
        x0 = along_lane
        n = len(self.control_points)
        S = sum(lengths)

        if x0 < 0:
            beta = x0
            return beta

        if x0 > S:
            beta = (n - 1.0) + (x0 - S)
            return beta

        if almost_equal(x0, S):
            beta = n - 1.0
            return beta

        assert 0 <= x0 < S, (x0, S)

        for i in range(n - 1):
            start_x = sum(lengths[:i])
            end_x = sum(lengths[: i + 1])
            if start_x <= x0 < end_x:
                beta = i + (x0 - start_x) / lengths[i]
                return beta

        assert False

    def draw_svg(self, drawing, g):
        assert isinstance(drawing, svgwrite.Drawing)
        glane = drawing.g()
        glane.attribs["class"] = "lane"

        cp1 = self.control_points[0]
        cp2 = self.control_points[-1]
        rel = relative_pose(cp1.as_SE2(), cp2.as_SE2())
        _, theta = geo.translation_angle_from_SE2(rel)

        delta = int(np.sign(theta))
        fill = {-1: "#577094", 0: "#709457", 1: "#947057"}[delta]

        points = self.lane_profile(points_per_segment=10)
        p = drawing.polygon(points=points, fill=fill, fill_opacity=0.5)

        glane.add(p)

        center_points = self.center_line_points(points_per_segment=10)
        center_points = [geo.translation_angle_from_SE2(_)[0] for _ in center_points]
        p = drawing.polyline(
            points=center_points, stroke=fill, fill="none", stroke_dasharray="0.02", stroke_width=0.01,
        )
        glane.add(p)

        g.add(glane)

        for x in self.control_points:
            q = x.asmatrix2d().m
            p1, _ = geo.translation_angle_from_SE2(q)
            delta_arrow = np.array([self.width / 4, 0])
            p2 = SE2_apply_R2(q, delta_arrow)
            gp = drawing.g()
            gp.attribs["class"] = "control-point"
            l = drawing.line(
                start=p1.tolist(), end=p2.tolist(), stroke="black", stroke_width=self.width / 20.0,
            )
            gp.add(l)
            c = drawing.circle(
                center=p1.tolist(),
                r=self.width / 8,
                fill="white",
                stroke="black",
                stroke_width=self.width / 20.0,
            )
            gp.add(c)
            g.add(gp)

    def center_point(self, beta: float) -> SE2value:
        n = len(self.control_points)
        i = int(np.floor(beta))

        if i < 0:
            q0 = self.control_points[0].asmatrix2d().m
            q1 = geo.SE2.multiply(q0, geo.SE2_from_translation_angle([0.1, 0], 0))
            alpha = beta

        elif i >= n - 1:
            # q0 = self.control_points[-2].asmatrix2d().m
            q0 = self.control_points[-1].asmatrix2d().m
            q1 = geo.SE2.multiply(q0, geo.SE2_from_translation_angle([0.1, 0], 0))
            alpha = beta - (n - 1)
        else:
            alpha = beta - i
            q0 = self.control_points[i].asmatrix2d().m
            q1 = self.control_points[i + 1].asmatrix2d().m
        q = SE2_interpolate(q0, q1, alpha)
        return q

    @memoized_reset
    def center_line_points(self, points_per_segment=5):
        n = len(self.control_points) - 1
        num = n * points_per_segment
        betas = np.linspace(0, n, num=num)
        res = []
        for beta in betas:
            q = self.center_point(beta)
            res.append(q)
        return res

    @memoized_reset
    def lane_profile(self, points_per_segment: int = 5):
        points_left = []
        points_right = []
        n = len(self.control_points) - 1
        num = n * points_per_segment
        betas = np.linspace(0, n, num=num)
        for beta in betas:
            q = self.center_point(beta)
            delta_left = np.array([0, self.width / 2])
            delta_right = np.array([0, -self.width / 2])
            points_left.append(SE2_apply_R2(q, delta_left))
            points_right.append(SE2_apply_R2(q, delta_right))

        return points_right + list(reversed(points_left))


new_contract("LaneSegment", LaneSegment)


def get_distance_two(q0, q1):
    g = geo.SE2.multiply(geo.SE2.inverse(q0), q1)
    v = geo.SE2.algebra_from_group(g)
    linear, angular = geo.linear_angular_from_se2(v)
    return np.linalg.norm(linear)



# coding=utf-8
import yaml
from comptests import comptest, run_module_tests, get_comptests_output_dir

# language=yaml
from duckietown_world import construct_map, draw_static, get_object_tree

map_yaml = """
objects:
- kind: sign_right_T_intersect
  attach:
    tile: [0, 0]
    slot: 2
    
    
tiles:
- [ asphalt, straight/S, asphalt]
- [ straight/E, 4way, straight/W]
- [ asphalt, straight/N, asphal]  
- [ asphalt, curve_right/W, straight/W]

tile_size: 0.61
"""


@comptest
def tag_positions():
    map_yaml_data = yaml.load(map_yaml, Loader=yaml.Loader)
    m = construct_map(map_yaml_data)
    print(get_object_tree(m, attributes=True))
    outdir = get_comptests_output_dir()
    draw_static(m, outdir)


if __name__ == "__main__":
    run_module_tests()



import numpy as np

import geometry as geo
from duckietown_world import SampledSequence
from duckietown_world.geo.transforms import SE2value
from duckietown_world.seqs.tsequence import SampledSequenceBuilder
from duckietown_world.world_duckietown.types import SE2v, se2v

from geometry import SE2


def get_velocities_from_sequence(s: SampledSequence[SE2v]) -> SampledSequence[se2v]:
    ssb = SampledSequenceBuilder[se2v]()
    ssb.add(0, geo.se2.zero())
    for i in range(1, len(s)):
        t0 = s.timestamps[i - 1]
        t1 = s.timestamps[i]
        q0 = s.values[i - 1]
        q1 = s.values[i]
        v = velocity_from_poses(t0, q0, t1, q1)
        if i == 0:
            ssb.add(t0, v)
        ssb.add(t1, v)
    return ssb.as_sequence()


def velocity_from_poses(t1: float, q1: SE2v, t2: float, q2: SE2v) -> se2v:
    delta = t2 - t1
    if not delta > 0:
        msg = f"invalid delta {delta}"
        raise ValueError(msg)

    x = SE2.multiply(SE2.inverse(q1), q2)
    xt = SE2.algebra_from_group(x)
    v = xt / delta
    return v


def relative_pose(base: SE2value, pose: SE2value) -> SE2value:
    assert isinstance(base, np.ndarray), base
    assert isinstance(pose, np.ndarray), pose
    return np.dot(np.linalg.inv(base), pose)



from OpenGL.GL import *
from OpenGL.GLUT import *
from OpenGL.raw.GLU import gluBuild2DMipmaps

from game1 import MapDistribution, create_map, sample_cars, get_background_image

w, h = 500, 500


def square():
    glBegin(GL_QUADS)
    glVertex2f(100, 100)
    glVertex2f(200, 100)
    glVertex2f(200, 200)
    glVertex2f(100, 200)
    glEnd()

a = 0

def iterate():
    global a
    a += 0.1
    glViewport(0, 0, 500, 500)
    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    glOrtho(0.0, 500, 0.0, 500+a, 0.0, 1.0)
    glMatrixMode(GL_MODELVIEW)
    glLoadIdentity()



def showScreen():
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
    glLoadIdentity()
    iterate()
    glColor3f(1.0, 0.0, 3.0)
    square()
    glutSwapBuffers()

import numpy as np
def get_texture(img: np.ndarray):
    h, w = img.shape[:2]

    glEnable(GL_TEXTURE_2D)
    textureID = glGenTextures(1)
    glPixelStorei(GL_UNPACK_ALIGNMENT, 4)
    glBindTexture(GL_TEXTURE_2D, textureID)
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0)
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0)
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, h, w,
                 0, GL_RGB, GL_UNSIGNED_BYTE, img.tobytes())
    return textureID

def go():
    glutInit()
    glutInitDisplayMode(GLUT_RGBA)
    glutInitWindowSize(500, 500)
    glutInitWindowPosition(0, 0)

    window_name = 'wn'
    md = MapDistribution(n=500, min_d=50, max_d=80, width_road=10)
    wm = create_map(md)
    N = 200
    N_min = 100
    cars = sample_cars(wm, n=N)
    bg = get_background_image(wm)


    textureID = get_texture(bg)

    # texture = glGenTextures(1)
    # glPixelStorei(GL_UNPACK_ALIGNMENT, 1)
    # glBindTexture(GL_TEXTURE_2D, texture)
    # glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
    # glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
    # glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
    # glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
    # gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB, h, w, GL_RGB, GL_UNSIGNED_BYTE, bg)

    #
    # glTexImage2D(GL_TEXTURE_2D,
    #              0,
    #              GL_RGB,
    #              w, h,
    #              0,
    #              GL_RGB,
    #              GL_UNSIGNED_BYTE,
    #              bg)

    it = 0

    frame = bg.copy()

    wind = glutCreateWindow("OpenGL Coding Practice")
    glutDisplayFunc(showScreen)
    glutIdleFunc(showScreen)


    # while True:
    #     it += 1
    #     update_cars(wm, cars, dt=Decimal(1), min_cars=N_min)
    #
    #     if it % 1 == 0:
    #
    #         ncars = 10
    #         pretties = []
    #         for k in list(cars)[:ncars]:
    #             if not cars[k].obs_history:
    #                 continue
    #             pretty = pretty_obs(cars[k].obs_history[-1])
    #             pretties.append(np.ones((pretty.shape[0], 5, 3), 'uint8') * 255)
    #             pretties.append(pretty)
    #
    #         rgb = np.hstack(tuple(pretties))
    #         cv2.imshow("one", rgb)
    #
    #         draw_cars(bg, frame, cars, erase=False)
    #
    #         cv2.imshow(window_name, frame)
    #         draw_cars(bg, frame, cars, erase=True)
    #         cv2.waitKey(1)

            # print(it)
    # cv2.destroyAllWindows()
    glutMainLoop()

if __name__ == '__main__':
    go()



import textwrap

import numpy as np

import geometry as geo
from duckietown_world.seqs import iterate_with_dt, SampledSequence, UndefinedAtTime
from duckietown_world.seqs.tsequence import SampledSequenceBuilder, Timestamp
from duckietown_world.world_duckietown import GetLanePoseResult, LanePose
from ..world_duckietown.utils import relative_pose
from .rule import Rule, RuleEvaluationContext, RuleEvaluationResult

__all__ = [
    "InDrivableLane",
    "DeviationFromCenterLine",
    "DeviationHeading",
    "DrivenLength",
    "DrivenLengthConsecutive",
    "SurvivalTime",
]


def integrate(sequence: SampledSequence[float]) -> SampledSequence[float]:
    """ Integrates with respect to time.
        That is, it multiplies the value with the Delta T. """
    if not sequence:
        msg = "Cannot integrate empty sequence."
        raise ValueError(msg)
    total = 0.0
    timestamps = []
    values = []
    for _ in iterate_with_dt(sequence):
        v0 = float(_.v0)
        dt = _.dt
        total += float(v0 * dt)

        timestamps.append(Timestamp(_.t0))
        values.append(total)

    return SampledSequence[float](timestamps, values)


def accumulate(sequence: SampledSequence[float]) -> SampledSequence[float]:
    """ Integrates with respect to time.
        Sums the values along the horizontal. """
    total = 0.0
    timestamps = []
    values = []
    for t, v in sequence:
        total += v

        timestamps.append(t)
        values.append(total)

    return SampledSequence[float](timestamps, values)


from typing import cast


class SurvivalTime(Rule):
    def evaluate(self, context: RuleEvaluationContext, result: RuleEvaluationResult):
        lane_pose_seq = context.get_lane_pose_seq()
        if len(lane_pose_seq) < 1:
            raise ValueError(lane_pose_seq)

        title = "Survival time"
        description = "Length of the episode."

        incremental = lane_pose_seq.transform_values(lambda _: 1.0, float)
        cumulative = integrate(incremental)
        total = cumulative.values[-1]

        result.set_metric(
            name=(),
            title=title,
            description=description,
            total=total,
            incremental=incremental,
            cumulative=cumulative,
        )


class DeviationFromCenterLine(Rule):
    def evaluate(self, context: RuleEvaluationContext, result: RuleEvaluationResult):

        interval = cast(SampledSequence, context.get_interval())
        lane_pose_seq = context.get_lane_pose_seq()

        timestamps = []
        values = []

        for i, timestamp in interval:
            try:
                name2lpr = lane_pose_seq.at(timestamp)
            except UndefinedAtTime:
                d = 0.0
            else:
                if name2lpr:
                    first = name2lpr[sorted(name2lpr)[0]]
                    assert isinstance(first, GetLanePoseResult)

                    lp = first.lane_pose
                    assert isinstance(lp, LanePose)

                    d = lp.distance_from_center
                else:
                    # no lp
                    d = 0.0

            values.append(d)
            timestamps.append(timestamp)

        sequence = SampledSequence[float](timestamps, values)

        if len(sequence) <= 1:
            cumulative = 0
            dtot = 0
        else:
            cumulative = integrate(sequence)
            dtot = cumulative.values[-1]

        title = "Deviation from center line"
        description = textwrap.dedent(
            """\
            This metric describes the amount of deviation from the center line.
        """
        )
        result.set_metric(
            name=(),
            total=dtot,
            incremental=sequence,
            title=title,
            description=description,
            cumulative=cumulative,
        )


class DeviationHeading(Rule):
    def evaluate(self, context: RuleEvaluationContext, result: RuleEvaluationResult):

        interval = cast(SampledSequence, context.get_interval())
        lane_pose_seq = context.get_lane_pose_seq()

        timestamps = []
        values = []

        for i, timestamp in interval:
            try:
                name2lpr = lane_pose_seq.at(timestamp)
            except UndefinedAtTime:
                d = 0.0
            else:
                if name2lpr:
                    first = name2lpr[sorted(name2lpr)[0]]
                    assert isinstance(first, GetLanePoseResult)

                    lp = first.lane_pose
                    assert isinstance(lp, LanePose)

                    d = np.abs(lp.relative_heading)
                else:
                    # no lp
                    d = 0.0

            values.append(d)
            timestamps.append(timestamp)

        sequence = SampledSequence[float](timestamps, values)
        if len(sequence) <= 1:
            cumulative = 0.0
            dtot = 0.0
        else:
            cumulative = integrate(sequence)
            dtot = cumulative.values[-1]

        # result.set_metric((), dtot, sequence, description, cumulative=cumulative)
        title = "Deviation from lane direction"
        description = textwrap.dedent(
            """\
            This metric describes the amount of deviation from the relative heading.
        """
        )
        result.set_metric(
            name=(),
            total=dtot,
            incremental=sequence,
            title=title,
            description=description,
            cumulative=cumulative,
        )


class InDrivableLane(Rule):
    def evaluate(self, context: RuleEvaluationContext, result: RuleEvaluationResult):
        interval = cast(SampledSequence, context.get_interval())
        lane_pose_seq = context.get_lane_pose_seq()

        timestamps = []
        values = []

        for i, timestamp in interval:
            try:
                name2lpr = lane_pose_seq.at(timestamp)
            except UndefinedAtTime:
                d = 1.0
            else:
                if name2lpr:
                    d = 0.0
                else:
                    # no lp
                    d = 1.0

            values.append(d)
            timestamps.append(timestamp)

        sequence = SampledSequence[float](timestamps, values)
        if len(sequence) <= 1:
            cumulative = 0
            dtot = 0
        else:
            cumulative = integrate(sequence)
            dtot = cumulative.values[-1]

        title = "Drivable areas"
        description = textwrap.dedent(
            """\
            This metric computes whether the robot was in a drivable area.
            
            Note that we check that the robot is in the lane in a correct heading 
            (up to 90deg deviation from the lane direction). 
        """
        )

        result.set_metric(
            name=(),
            total=dtot,
            incremental=sequence,
            title=title,
            description=description,
            cumulative=cumulative,
        )


class DrivenLength(Rule):
    def evaluate(self, context: RuleEvaluationContext, result: RuleEvaluationResult):
        interval = context.get_interval()
        lane_pose_seq = context.get_lane_pose_seq()
        ego_pose_sequence = context.get_ego_pose_global()

        driven_any_builder = SampledSequenceBuilder[float]()
        driven_lanedir_builder = SampledSequenceBuilder[float]()

        for idt in iterate_with_dt(interval):
            t0, t1 = idt.v0, idt.v1  # not v
            try:
                name2lpr = lane_pose_seq.at(t0)

                p0 = ego_pose_sequence.at(t0).as_SE2()
                p1 = ego_pose_sequence.at(t1).as_SE2()
            except UndefinedAtTime:
                dr_any = dr_lanedir = 0.0

            else:
                prel = relative_pose(p0, p1)
                translation, _ = geo.translation_angle_from_SE2(prel)
                dr_any = np.linalg.norm(translation)

                if name2lpr:

                    ds = []
                    for k, lpr in name2lpr.items():
                        assert isinstance(lpr, GetLanePoseResult)
                        c0 = lpr.center_point
                        ctas = geo.translation_angle_scale_from_E2(c0.asmatrix2d().m)
                        c0_ = geo.SE2_from_translation_angle(ctas.translation, ctas.angle)
                        prelc0 = relative_pose(c0_, p1)
                        tas = geo.translation_angle_scale_from_E2(prelc0)

                        # otherwise this lane should not be reported
                        # assert tas.translation[0] >= 0, tas
                        ds.append(tas.translation[0])

                    dr_lanedir = max(ds)
                else:
                    # no lp
                    dr_lanedir = 0.0

            driven_any_builder.add(t0, dr_any)
            driven_lanedir_builder.add(t0, dr_lanedir)

        driven_any_incremental = driven_any_builder.as_sequence()

        driven_any_cumulative = accumulate(driven_any_incremental)

        if len(driven_any_incremental) <= 1:
            total = 0
        else:
            total = driven_any_cumulative.values[-1]

        title = "Distance"
        description = textwrap.dedent(
            """\
            This metric computes how far the robot drove.
        """
        )

        result.set_metric(
            name=("driven_any",),
            total=total,
            incremental=driven_any_incremental,
            title=title,
            description=description,
            cumulative=driven_any_cumulative,
        )
        title = "Lane distance"

        driven_lanedir_incremental = driven_lanedir_builder.as_sequence()
        driven_lanedir_cumulative = accumulate(driven_lanedir_incremental)

        if len(driven_lanedir_incremental) <= 1:
            total = 0
        else:
            total = driven_lanedir_cumulative.values[-1]

        description = textwrap.dedent(
            """\
            This metric computes how far the robot drove
            **in the direction of the lane**.
        """
        )
        result.set_metric(
            name=("driven_lanedir",),
            total=total,
            incremental=driven_lanedir_incremental,
            title=title,
            description=description,
            cumulative=driven_lanedir_cumulative,
        )


class DrivenLengthConsecutive(Rule):
    def evaluate(self, context: RuleEvaluationContext, result: RuleEvaluationResult):
        interval = context.get_interval()
        lane_pose_seq = context.get_lane_pose_seq()
        ego_pose_sequence = context.get_ego_pose_global()

        timestamps = []
        driven_any = []
        driven_lanedir = []

        tile_fqn2lane_fqn = {}
        for idt in iterate_with_dt(interval):
            t0, t1 = idt.v0, idt.v1  # not v
            try:
                name2lpr = lane_pose_seq.at(t0)

                p0 = ego_pose_sequence.at(t0).as_SE2()
                p1 = ego_pose_sequence.at(t1).as_SE2()
            except UndefinedAtTime:
                dr_any = dr_lanedir = 0.0

            else:
                prel = relative_pose(p0, p1)
                translation, _ = geo.translation_angle_from_SE2(prel)
                dr_any = np.linalg.norm(translation)

                if name2lpr:

                    ds = []
                    for k, lpr in name2lpr.items():
                        if lpr.tile_fqn in tile_fqn2lane_fqn:
                            if lpr.lane_segment_fqn != tile_fqn2lane_fqn[lpr.tile_fqn]:
                                # msg = 'Backwards detected'
                                # print(msg)
                                continue

                        tile_fqn2lane_fqn[lpr.tile_fqn] = lpr.lane_segment_fqn

                        assert isinstance(lpr, GetLanePoseResult)
                        c0 = lpr.center_point
                        ctas = geo.translation_angle_scale_from_E2(c0.asmatrix2d().m)
                        c0_ = geo.SE2_from_translation_angle(ctas.translation, ctas.angle)
                        prelc0 = relative_pose(c0_, p1)
                        tas = geo.translation_angle_scale_from_E2(prelc0)

                        # otherwise this lane should not be reported
                        # assert tas.translation[0] >= 0, tas
                        ds.append(tas.translation[0])

                    dr_lanedir = max(ds) if ds else 0.0
                else:
                    # no lp
                    dr_lanedir = 0.0

            driven_any.append(dr_any)
            driven_lanedir.append(dr_lanedir)
            timestamps.append(t0)

        title = "Consecutive lane distance"

        driven_lanedir_incremental = SampledSequence[float](timestamps, driven_lanedir)
        driven_lanedir_cumulative = accumulate(driven_lanedir_incremental)

        if len(driven_lanedir_incremental) <= 1:
            total = 0
        else:
            total = driven_lanedir_cumulative.values[-1]
        description = textwrap.dedent(
            """\
            This metric computes how far the robot drove **in the direction of the correct lane**,
            discounting whenever it was driven in the wrong direction with respect to the start.
        """
        )
        result.set_metric(
            name=("driven_lanedir_consec",),
            total=total,
            incremental=driven_lanedir_incremental,
            title=title,
            description=description,
            cumulative=driven_lanedir_cumulative,
        )



import random
from dataclasses import dataclass, replace
from decimal import Decimal
from typing import Dict, List, Set, Tuple, TypeVar

import cv2


@dataclass
class GameDescription:
    half_width: int  # 5


PL_SELF = 'self'
PL_NATURE = 'nature'
PL_OTHER = 'nature'

# what choices

X = TypeVar('X')


@dataclass
class DecPose:
    x: Decimal
    y: Decimal
    theta: Decimal
    vx: Decimal


@dataclass
class CarProperties:
    width: Decimal
    length: Decimal
    max_vx: Decimal
    max_accel: Decimal
    max_brake: Decimal
    max_side_by_vx: List[Decimal]
    field_of_view: Decimal


@dataclass
class CarState:
    pose: DecPose  # with respect to the lane
    turn_signal: int  # -1 left, 0 none, +1 right
    front_light: bool
    brake_light: bool


#  v += gas - brake * 2
#
#  v = 0
#  v = 1
#  v = 2
#      move: +1, 0, -1
#  v = 3
#      move: +2, +1, 0, -1, -2


@dataclass
class CarCommands:
    brake: int  # between 0 and 1
    gas: int  # between 0 and 1
    lateral: int

    turn_signal: int
    front_light: bool


@dataclass
class GameState:
    pass
    # your_pose: CarState
    # their_pose: Set[CarState]


import numpy as np


@dataclass
class WorldMap:
    obstacles: np.ndarray
    horizontal: np.ndarray
    vertical: np.ndarray
    horizontal_roads: List[int]
    vertical_roads: List[int]


def sample_pose(wm: WorldMap) -> Tuple[int, int]:
    n, q = wm.obstacles.shape[0:2]
    B = 50
    while True:
        i = random.randint(B, n - B - 1)
        j = random.randint(B, q - B - 1)
        if not wm.obstacles[i, j]:
            return i, j


@dataclass
class CarObservations:
    local_map: np.ndarray
    local_obs: np.ndarray
    visible: np.ndarray
    relative: np.ndarray

    def __post_init__(self):
        s1 = self.local_map.shape
        s2 = self.local_obs.shape
        s3 = self.visible.shape
        s4 = self.relative.shape[:2]
        assert s1 == s2 == s3 == s4, (s1, s2, s3, s4)


@dataclass
class Car:
    cp: CarProperties
    cs: CarState
    policy: 'Policy'
    obs_history: List[CarObservations]


class Policy:
    def get_action(self, obs_history: List[CarObservations], cp: CarProperties) -> CarCommands:
        pass


class SimplePolicy(Policy):
    def get_action(self, obs_history: List[CarObservations], cp: CarProperties) -> CarCommands:
        lateral = random.randint(-1, +1)
        gas = random.random() > 0.4
        brake = random.random() > 0.8
        turn_signal = random.randint(-1, +1)
        c = CarCommands(brake=brake, gas=gas, lateral=lateral, turn_signal=turn_signal, front_light=False)
        return c


class AvoidPolicy(Policy):
    def get_action(self, obs_history: List[CarObservations], cp: CarProperties) -> CarCommands:
        last = obs_history[-1]
        dx = last.relative[:, :, 0]
        dy = last.relative[:, :, 1]
        # d = np.hypot(dx, dy)
        theta = np.arctan2(dy, dx)

        close1 = np.rad2deg(np.abs(theta)) < 50

        close2 = np.logical_and(dx>2, dx < 20)
        close = np.logical_and(close1, close2)
        if np.any(last.local_obs[close]):
            gas = 0
            brake = 1
        else:

            gas = 1
            brake = 0
        lateral = 0
        # lateral = random.randint(-1, +1)
        # gas = random.random() > 0.4
        # brake = random.random() > 0.8
        turn_signal = random.randint(-1, +1)
        c = CarCommands(brake=brake, gas=gas, lateral=lateral, turn_signal=turn_signal, front_light=False)
        return c


def sample_car(wm: WorldMap) -> Car:
    i, j = sample_pose(wm)
    if wm.horizontal[i, j] > 0:
        theta = Decimal(np.pi / 2)
    else:
        theta = Decimal(0)
    pick = np.random.rand() > 0.5
    if pick:
        theta = theta + Decimal(np.pi)

    pose = DecPose(x=Decimal(i), y=Decimal(j), theta=Decimal(theta),
                   vx=Decimal(1))
    field_of_view = np.random.randint(10, 50)
    cs = CarState(pose=pose, turn_signal=0, front_light=False, brake_light=False)
    cp = CarProperties(width=Decimal(2), length=Decimal(3),
                       max_vx=Decimal(3),
                       max_accel=Decimal(1),
                       max_brake=Decimal(2),
                       field_of_view=Decimal(field_of_view),
                       max_side_by_vx=[Decimal(0), Decimal(0), Decimal(1), Decimal(2)])

    if random.random() > 0.4:
        policy = AvoidPolicy()
    else:
        policy = SimplePolicy()
    car = Car(cp, cs, policy, [])
    return car


def sample_cars(wm: WorldMap, n: int) -> Dict[str, Car]:
    cars: Dict[str, Car] = {}
    for a in range(n):
        name = str(a)
        cars[name] = sample_car(wm)
    return cars


def zoom_d(a, z: int):
    return np.kron(a, np.ones((z, z), 'uint8'))


def get_background_image(wm: WorldMap):
    h, w = wm.obstacles.shape
    rgb = np.zeros((h, w, 3), 'uint8')
    rgb[wm.obstacles] = 255
    return rgb


def get_car_fov(theta: Decimal) -> Tuple[Tuple[int, int], Tuple[int, int]]:
    fovX = -3, +10
    fovY = -2, +6

    fovX = -20, +30
    fovY = -15, +15
    return get_area(theta, fovX, fovY)


def get_car_footprint(theta: Decimal) -> Tuple[Tuple[int, int], Tuple[int, int]]:
    fovX = -1, +2
    fovY = -1, +1
    return get_area(theta, fovX, fovY)


def get_area(theta: Decimal, fovX, fovY) -> Tuple[Tuple[int, int], Tuple[int, int]]:
    nx, ny = int(np.cos(float(theta))), int(np.sin(float(theta)))

    def swap(x) -> Tuple[int, int]:
        return -x[1], -x[0]

    if (nx, ny) == (1, 0):
        return fovX, fovY
    if (nx, ny) == (-1, 0):
        return swap(fovX), fovY
    if (nx, ny) == (0, 1):
        return fovY, fovX
    if (nx, ny) == (0, -1):
        return fovY, swap(fovX)
    assert False, (theta, nx, ny)


def compute_observations(wm: WorldMap, cars: Dict[str, Car]):
    h, w = wm.obstacles.shape
    moving_obstacles = np.zeros((h, w), int)
    shape = wm.obstacles.shape

    for k, car in cars.items():
        cs: CarState = car.cs
        i = int(cs.pose.x)
        j = int(cs.pose.y)
        fpx, fpy = get_car_footprint(car.cs.pose.theta)
        i1 = i + fpx[0]
        i2 = i + fpx[1]
        j1 = j + fpy[0]
        j2 = j + fpy[1]
        i1, i2, j1, j2 = clip(shape, i1, i2, j1, j2)

        assert 0 <= j1 <= j2
        assert 0 <= i1 <= i2
        moving_obstacles[i1:i2, j1:j2] = 1

    fpx, fpy = get_car_fov(Decimal(0))
    the_is = list(range(fpx[0], fpx[1]))
    the_js = list(range(fpy[0], fpy[1]))
    s2 = (len(the_is), len(the_js))
    relative_i = np.zeros(s2, int)
    relative_j = np.zeros(s2, int)
    for m, a in enumerate(the_is):
        relative_i[m, :] = a
    for m, b in enumerate(the_js):
        relative_j[:, m] = b
    ds2 = relative_i * relative_i + relative_j * relative_j

    relative = np.dstack((relative_i, relative_j))

    res = {}
    for k, car in cars.items():
        cs: CarState = car.cs
        i = int(cs.pose.x)
        j = int(cs.pose.y)
        fpx, fpy = get_car_fov(car.cs.pose.theta)

        i1 = i + fpx[0]
        i2 = i + fpx[1]
        j1 = j + fpy[0]
        j2 = j + fpy[1]
        i1, i2, j1, j2 = clip(shape, i1, i2, j1, j2)
        assert 0 <= j1 <= j2
        assert 0 <= i1 <= i2
        sense_obstacles = wm.obstacles[i1:i2, j1:j2].copy()
        sense_cars = moving_obstacles[i1:i2, j1:j2].copy()

        # print(ds.flatten())
        outside = ds2 > car.cp.field_of_view * car.cp.field_of_view
        visible = np.logical_not(outside)

        sense_obstacles = rotate(sense_obstacles, car.cs.pose.theta)
        sense_cars = rotate(sense_cars, car.cs.pose.theta)

        sense_obstacles[outside] = 0
        sense_cars[outside] = 0

        obs = CarObservations(sense_obstacles, sense_cars, visible, relative)
        res[k] = obs
    return res


def rotate(x, theta):
    nx, ny = int(np.cos(float(theta))), int(np.sin(float(theta)))
    if (nx, ny) == (1, 0):
        return x
    if (nx, ny) == (-1, 0):
        return np.rot90(np.rot90(x))
    if (nx, ny) == (0, -1):
        return np.rot90(x)
    if (nx, ny) == (0, 1):
        return np.rot90(np.rot90(np.rot90(x)))
    assert False, theta


def clip(shape, i1, i2, j1, j2):
    i1 = max(0, i1)
    i2 = min(shape[0] - 1, i2)
    j1 = max(0, j1)
    j2 = min(shape[1] - 1, j2)
    return i1, i2, j1, j2


def draw_cars(background, rgb, cars: Dict[str, Car], erase=False):


    shape = background.shape[:2]
    for k, car in cars.items():
        cs: CarState = car.cs
        nx = int(np.cos(float(car.cs.pose.theta)))
        ny = int(np.sin(float(car.cs.pose.theta)))

        i = int(cs.pose.x)
        j = int(cs.pose.y)
        fpx, fpy = get_car_footprint(car.cs.pose.theta)
        i1 = i + fpx[0]
        i2 = i + fpx[1]
        j1 = j + fpy[0]
        j2 = j + fpy[1]
        i1, i2, j1, j2 = clip(shape, i1, i2, j1, j2)
        assert 0 <= j1 <= j2
        assert 0 <= i1 <= i2

        colors = {
            (0,1): (255,0,0),
            (0,-1):(0,255,0),
            (1, 0):(0,0,255),
            (-1, 0):(120,0,120),
        }
        color = colors[(nx,ny)]
        if erase:
            rgb[i1:i2, j1:j2, :] = background[i1:i2, j1:j2, :]
        else:
            rgb[i1:i2, j1:j2, :] = color

    return rgb


def compute_collisions(wm: WorldMap, cars: Dict[str, Car]) -> \
    Tuple[Set[Tuple[str, str]], Set[str]]:
    c = np.zeros(wm.obstacles.shape, dtype=int)
    c.fill(-1)
    names = list(cars)
    collisions = set()
    outside = set()
    for m, k in enumerate(names):
        car = cars[k]

        cs: CarState = car.cs
        i = int(cs.pose.x)
        j = int(cs.pose.y)
        w = 1  # int(car.cp.width / 2)
        l = 1  # int(car.cp.length / 2)

        obstacles = wm.obstacles[i - w:i + w + 1, j - l:j + l + 1]
        if np.any(obstacles):
            outside.add(names[m])

        ground = c[i - w:i + w + 1, j - l:j + l + 1]
        if len(ground.flatten()) == 0:
            continue

        already = np.max(ground.flatten())
        if already >= 0:
            collisions.add((names[already], names[m]))

        c[i - w:i + w, j - l:j + l] = m

    return collisions, outside


@dataclass
class MapDistribution:
    n: int
    min_d: int
    max_d: int
    width_road: int


def create_map(md: MapDistribution) -> WorldMap:
    n = md.n
    min_d = md.min_d
    max_d = md.max_d
    width_road = md.width_road
    x = np.zeros((n, n), bool)
    vertical = np.zeros((n, n), int)
    vertical.fill(-1)
    horizontal = np.zeros((n, n), int)
    horizontal.fill(-1)
    x.fill(1)

    i_s = get_roads(n, min_d, max_d, width_road)
    j_s = get_roads(n, min_d, max_d, width_road)
    w = width_road

    for i in i_s:
        x[i - w:i + w, :] = False
        horizontal[i - w:i + w, :] = i
    for j in j_s:
        x[:, j - w:j + w] = False
        vertical[:, j - w:j + w] = j

    return WorldMap(obstacles=x, horizontal=horizontal, vertical=vertical,
                    horizontal_roads=i_s, vertical_roads=j_s)


def get_roads(n: int, min_d: int, max_d: int, width: int) -> List[int]:
    i = width
    res = []

    def sample() -> int:
        dx = random.randint(min_d, max_d)
        return dx

    while True:
        dx = sample()
        i += dx

        if i >= n - width:
            break
        res.append(i)
        i += width
    return res


def reset_coords(wm: WorldMap, cp1: DecPose, cp2: DecPose) -> DecPose:
    x = cp2.x
    y = cp2.y
    vx = cp2.vx
    theta = cp2.theta
    B = 35
    H, W = wm.obstacles.shape
    if x < B:
        x = H - B
    if y < B:
        y = W - B

    if x > H - B:
        x = B

    if y > W - B:
        y = B
    return DecPose(x=x, y=y, vx=vx, theta=theta)


def update_pose(wm: WorldMap, cp: DecPose, cprop: CarProperties, commands: CarCommands, dt: Decimal) -> DecPose:
    assert dt == 1
    nx = Decimal(np.cos(float(cp.theta)))
    ny = Decimal(np.sin(float(cp.theta)))

    assert 0 <= commands.brake <= 1
    assert 0 <= commands.gas <= 1
    vx = cp.vx + cprop.max_accel * commands.gas - cprop.max_brake * commands.brake
    vx = max(0, vx)
    vx = min(cprop.max_vx, vx)
    max_side = cprop.max_side_by_vx[int(cp.vx)]
    side = max(min(commands.lateral, max_side), -max_side)

    dx = cp.vx * nx - side * ny
    dy = cp.vx * ny + side * nx

    x = cp.x + dx
    y = cp.y + dy
    theta = cp.theta
    cp2 = DecPose(x=x, y=y, vx=vx, theta=theta)
    return reset_coords(wm, cp, cp2)


def update(wm: WorldMap, cs: CarState, cprop: CarProperties, commands: CarCommands, dt: Decimal) -> CarState:
    pose2 = update_pose(wm, cs.pose, cprop, commands, dt)
    # print(cs.pose, pose2)
    return replace(cs, pose=pose2)


carnames = 0


def update_cars(wm: WorldMap, cars: Dict[str, Car], dt: Decimal, min_cars: int):
    collisions, outside = compute_collisions(wm, cars)
    for k1, k2 in collisions:
        cars.pop(k1, None)
        cars.pop(k2, None)
    for k in outside:
        cars.pop(k, None)

    obs_per_car = compute_observations(wm, cars)
    for k, v in cars.items():
        obs = obs_per_car[k]
        v.obs_history.append(obs)
        commands = v.policy.get_action(v.obs_history, v.cp)
        cs = update(wm, v.cs, v.cp, commands, dt)
        v.cs = cs

    toadd = max(0, min_cars - len(cars))
    global carnames
    for i in range(toadd):
        car = sample_car(wm)
        name = str('new%s' % carnames)
        carnames += 1
        cars[name] = car


def pretty_obs(obs: CarObservations):
    a = obs.local_map.astype('uint8') * 255
    b = obs.local_obs.astype('uint8') * 255
    c = np.logical_not(obs.visible).astype('uint8') * 128

    # rgb = np.zeros((a.shape[0], a.shape[1], 3), 'uint8')
    # rgb[obs.local_map, 0] = 255
    # rgb[obs.local_obs, 1] = 255
    # rgb[obs.local_obs, 1] = 255
    # close = obs.relative[:, :, 0] > 0
    # a =b =c =close.astype('uint8')* 128
    bgr = np.dstack((a, b, c))

    bgr = np.rot90(np.rot90(bgr))
    return bgr
    # return np.vstack((a, b))


def go():
    window_name = 'wn'
    md = MapDistribution(n=500, min_d=50, max_d=80, width_road=10)
    wm = create_map(md)
    N = 200
    N_min = 100
    cars = sample_cars(wm, n=N)

    it = 0
    bg = get_background_image(wm)
    frame = bg.copy()
    while True:
        it += 1
        update_cars(wm, cars, dt=Decimal(1), min_cars=N_min)

        if it % 1 == 0:

            ncars = 10
            pretties = []
            for k in list(cars)[:ncars]:
                if not cars[k].obs_history:
                    continue
                pretty = pretty_obs(cars[k].obs_history[-1])
                pretties.append(np.ones((pretty.shape[0], 5, 3), 'uint8') * 255)
                pretties.append(pretty)

            rgb = np.hstack(tuple(pretties))
            cv2.imshow("one", rgb)

            draw_cars(bg, frame, cars, erase=False)

            cv2.imshow(window_name, frame)
            draw_cars(bg, frame, cars, erase=True)
            cv2.waitKey(1)

            print(it)
    cv2.destroyAllWindows()


if __name__ == '__main__':
    go()



# coding=utf-8
import argparse
import json
import os
import sys

from duckietown_world import logger
from .misc import draw_static

__all__ = ["draw_maps_main"]


def draw_maps_main(args=None):
    from duckietown_world.world_duckietown import list_maps, load_map

    if args is None:
        args = sys.argv[1:]
    parser = argparse.ArgumentParser()
    parser.add_argument("--output", help="output dir", default="out-draw_maps")
    parser.add_argument("map_names", nargs=argparse.REMAINDER)
    parsed = parser.parse_args(args)

    output = parsed.output

    if parsed.map_names:
        map_names = parsed.map_names
    else:
        map_names = list_maps()
    logger.info("Drawing the maps %s." % ", ".join(map_names))

    for map_name in map_names:
        duckietown_map = load_map(map_name)
        out = os.path.join(output, map_name)

        draw_map(out, duckietown_map)

        y = duckietown_map.as_json_dict()
        fn = os.path.join(out, "map.json")
        with open(fn, "w") as f:
            f.write(json.dumps(y, indent=4))
        # print('written to %s' % fn)


def draw_map(output: str, duckietown_map: "DuckietownMap") -> None:
    from duckietown_world.world_duckietown import DuckietownMap

    if not os.path.exists(output):
        os.makedirs(output)
    assert isinstance(duckietown_map, DuckietownMap)

    fns = draw_static(duckietown_map, output_dir=output, pixel_size=(640, 640), area=None)
    for fn in fns:
        logger.info("Written to %s" % fn)


if __name__ == "__main__":
    draw_maps_main()



# coding=utf-8
__all__ = ["Region", "EmptyRegion"]


class Region(object):
    def contains(self, p):
        """ Return true if it contains the 2D point p. """


class EmptyRegion(Region):
    def contains(self, p):
        return False



import os

import yaml
from duckietown_serialization_ds1 import Serializable
from PIL import Image
from six import BytesIO

from duckietown_world import logger, PlacedObject
from duckietown_world.utils import memoized_reset

# DEFAULT_FAMILY = '36h11'


__all__ = ["get_apriltagsDB_raw", "get_sign_type_from_tag_id"]


class TagInstance(Serializable):
    def __init__(self, tag_id, family, size):
        self.tag_id = tag_id
        self.family = family
        self.size = size

        texture = "tag%s_%05d.png" % (family.replace("h", "_"), tag_id)
        try:
            from duckietown_world.world_duckietown.map_loading import get_texture_file

            self.fn = get_texture_file(texture)
        except KeyError:
            msg = "Cannot find april tag image for %s" % texture
            logger.warning(msg)
            self.fn = None

    def draw_svg(self, drawing, g):
        # L = self.size
        # rect = drawing.rect(insert=insert,(-L / 2, -L / 2),
        #                     size=(L, L),
        #                     fill="white",
        #                     # style='opacity:0.4',
        #                     stroke_width="0.005",
        #                     stroke="black", )
        # g.add(rect)
        if self.fn:
            with Image.open(self.fn) as _:
                image = _.convert("RGB")
            image = image.resize((64, 64), resample=Image.NEAREST)
            out = BytesIO()
            image.save(out, format="png")
            texture = out.getvalue()

            from duckietown_world import data_encoded_for_src

            href = data_encoded_for_src(texture, "image/png")
            T = self.size
            insert = (-T / 2, -T / 2)

            img = drawing.image(
                href=href,
                size=(T, T),
                insert=insert,
                style="transform: rotate(90deg) scaleX(-1)  rotate(-90deg) ",
            )
            img.attribs["class"] = "tile-textures"
            g.add(img)


class FloorTag(PlacedObject):
    def __init__(self, tag=None, **kwargs):
        PlacedObject.__init__(self, **kwargs)
        self.tag = tag

    def draw_svg(self, drawing, g):
        self.tag.draw_svg(drawing, g)


@memoized_reset
def get_apriltagsDB_raw():
    abs_path_module = os.path.realpath(__file__)
    module_dir = os.path.dirname(abs_path_module)

    fn = os.path.join(module_dir, "../data/apriltagsDB.yaml")
    with open(fn) as _:
        d = _.read()
    return yaml.load(d)


def get_sign_type_from_tag_id(tag_id):
    tag_id = int(tag_id)
    db = get_apriltagsDB_raw()
    # logger.debug(set(_['traffic_sign_type'] for _ in db))
    # logger.debug(set(_['tag_type'] for _ in db))

    #
    #       street_name: BLUM ST.
    #       vehicle_name:
    #       traffic_sign_type:

    from .other_objects import get_canonical_sign_name

    for entry in db:
        if entry["tag_id"] == tag_id:
            tstype = entry["traffic_sign_type"]
            return get_canonical_sign_name(tstype)
    raise KeyError(tag_id)



# coding=utf-8

from comptests import comptest, run_module_tests, get_comptests_output_dir
from duckietown_world import get_object_tree

from duckietown_world.svg_drawing import draw_static
from duckietown_world.world_duckietown.map_loading import load_map
from duckietown_world.world_duckietown.segmentify import get_skeleton_graph


@comptest
def lane_pose_test1():
    outdir = get_comptests_output_dir()

    dm = load_map("udem1")

    print(get_object_tree(dm, attributes=True))

    res = get_skeleton_graph(dm)

    # area = RectangularArea((0, 0), (3, 3))
    draw_static(res.root2, outdir + "/root2")

    # draw_static(dm, outdir + '/orig')

    print(get_object_tree(res.root2, attributes=True))


if __name__ == "__main__":
    run_module_tests()



# coding=utf-8

from comptests import comptest, run_module_tests, get_comptests_output_dir

from duckietown_serialization_ds1 import Serializable
from duckietown_world import list_maps
from duckietown_world.geo import PlacedObject, SE2Transform
from duckietown_world.seqs import Constant
from duckietown_world.seqs.tsequence import SampledSequenceBuilder
from duckietown_world.world_duckietown import create_map
from duckietown_world.world_duckietown.map_loading import load_map


@comptest
def wb1():
    outdir = get_comptests_output_dir()
    root = PlacedObject()
    tile_map = create_map(H=3, W=3)

    world = PlacedObject()
    root.set_object("world", world)

    placement = Constant[SE2Transform](SE2Transform.identity())

    world.set_object("map1", tile_map, ground_truth=placement)

    ego = PlacedObject()
    world_coordinates = Constant[SE2Transform](SE2Transform([0, 0], 0))

    world.set_object("ego", ego, ground_truth=world_coordinates)

    d = root.as_json_dict()
    # print(json.dumps(DW.root.as_json_dict(), indent=4))
    # print(yaml.safe_dump(d, default_flow_style=False))
    # print('------')
    r1 = Serializable.from_json_dict(d)
    # print('read: %s' % r1)
    d1 = r1.as_json_dict()
    # print(yaml.safe_dump(d1, default_flow_style=False))
    # assert d == d1


@comptest
def wb2():
    root = PlacedObject()

    for map_name in list_maps():
        tm = load_map(map_name)
        root.set_object(map_name, tm)

    d = root.as_json_dict()
    # print(json.dumps(d, indent=4))
    # print(yaml.safe_dump(d, default_flow_style=False))

    # print('------')
    r1 = Serializable.from_json_dict(d)
    d1 = r1.as_json_dict()
    # print(yaml.safe_dump(d1, default_flow_style=False))
    # assert d == d1


@comptest
def sb1():
    a = SampledSequenceBuilder[float]()
    a.add(0.0, 0.0)


if __name__ == "__main__":
    run_module_tests()



# coding=utf-8
from typing import List

from duckietown_world import GenericSequence, PlacedObject, TransformSequence

__all__ = [
    "ChooseTime",
    # 'RemoveVariable',
    # 'RemoveStatic',
    "get_sampling_points",
]

from duckietown_world.seqs.tsequence import Timestamp
from zuper_commons.types import ZValueError


class ChooseTime(object):
    def __init__(self, t):
        self.t = t

    def __call__(self, ob):
        # if isinstance(ob, Sequence):
        if hasattr(ob, "at"):
            ob = ob.at(self.t)
            return ob
        else:
            return ob


#
#
# class RemoveVariable(object):
#     def __init__(self):
#         pass
#
#     def __call__(self, ob):
#         if isinstance(ob, PlacedObject):
#             toremove = []
#             for sr_name, sr in ob.spatial_relations.items():
#                 if not sr.a and isinstance(sr.transform, Sequence):
#                     toremove.append(sr.b)
#             return remove_stuff_and_children(ob, toremove)
#
#         else:
#             return ob

#
# class RemoveStatic(object):
#     def __init__(self):
#         pass
#
#     def __call__(self, ob):
#         if isinstance(ob, PlacedObject):
#             toremove = []
#             for sr_name, sr in ob.spatial_relations.items():
#                 # print(sr_name, sr)
#                 if not sr.a and isinstance(sr.transform, Sequence) and is_static(sr.transform):
#                     # print('removing')
#                     toremove.append(sr.b)
#             if toremove:
#                 print('removing %s' % toremove)
#             return remove_stuff_and_children(ob, toremove)
#         else:
#             return ob


def is_static(transform):
    if isinstance(transform, GenericSequence):
        return False
    if isinstance(transform, TransformSequence):
        return all(is_static(_) for _ in transform.transforms)
    return True


# def remove_stuff_and_children(ob, toremove):
#     assert isinstance(ob, PlacedObject)
#     children = {}
#     for child_name, child in ob.children.items():
#         if child_name in toremove:
#             pass
#         else:
#             children[child_name] = child
#     srs = {}
#     for sr_name, sr in ob.spatial_relations.items():
#         if sr.b[0] in toremove:
#             pass
#         else:
#             srs[sr_name] = sr
#     return PlacedObject(children=children, spatial_relations=srs)

#
# class Flatten(object):
#     def __init__(self):
#         pass
#
#     def __call__(self, ob):
#         if isinstance(ob, Transform):
#             return ob.asmatrix2d()
#         else:
#             return ob


def get_sampling_points(ob0: PlacedObject) -> List[Timestamp]:
    points = set()

    def f(ob: PlacedObject) -> PlacedObject:
        # print(ob)
        if hasattr(ob, "get_sampling_points"):
            if not isinstance(ob, GenericSequence):
                raise ZValueError(ob=ob)
            sp = ob.get_sampling_points()
            if sp == GenericSequence.CONTINUOUS:
                pass
            else:
                points.update(sp)
        # else:
        #     logger.info(f'not a seq: {type(ob)}')
        return ob

    ob0.filter_all(f)
    return sorted(points)



# coding=utf-8
from .other_objects import *
from .tile import *
from .tile_coords import *
from .tile_map import *
from .traffic_light import *

from .map_loading import *
from .platform_dynamics import *
from .integrator2d import *
from .generic_kinematics import *
from .lane_segment import *
from .duckietown_map import *
from .duckiebot import *
from .transformations import *
from .segmentify import *
from .pwm_dynamics import *



from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from dataclasses import dataclass
from typing import Dict, Optional, Tuple, Union

from duckietown_serialization_ds1 import Serializable
from duckietown_world import LanePose
from duckietown_world.geo import PlacedObject, SE2Transform
from duckietown_world.seqs import SampledSequence
from duckietown_world.seqs.tsequence import Timestamp
from duckietown_world.svg_drawing.misc import TimeseriesPlot
from zuper_commons.types import check_isinstance

__all__ = ["RuleEvaluationContext", "RuleEvaluationResult", "Rule", "evaluate_rules", 'EvaluatedMetric',
           'make_timeseries']


@dataclass
class RuleEvaluationContext:
    interval: SampledSequence[Timestamp]
    world: PlacedObject
    ego_name: str
    lane_pose_seq: SampledSequence[LanePose]
    pose_seq: SampledSequence[SE2Transform]

    def get_interval(self) -> SampledSequence[Timestamp]:
        """ Returns the interval over which to evaluate the rule. """
        return self.interval

    def get_world(self) -> PlacedObject:
        """ Returns the world object. """
        return self.world

    def get_ego_name(self) -> str:
        """ Returns the name of the ego-vehicle
            as an object in the hierarchy """
        return self.ego_name

    def get_lane_pose_seq(self) -> SampledSequence[LanePose]:
        """ Returns the lane pose result sequence.
            At each timestamp a possibly empty dict of index -> LanePoseResult """
        return self.lane_pose_seq

    def get_ego_pose_global(self) -> SampledSequence[SE2Transform]:
        """ Returns the global pose of the vehicle. """
        return self.pose_seq


class EvaluatedMetric(Serializable):
    total: float
    incremental: float
    cumulative: float
    description: str
    title: str

    def __init__(self, total, incremental, description, cumulative, title):
        self.total = float(total)
        self.title = title
        self.incremental = incremental
        self.cumulative = cumulative
        self.description = description

    def __repr__(self):
        return "EvaluatedMetric(%s, %s)" % (self.title, self.total)


class RuleEvaluationResult:
    metrics: Dict[str, EvaluatedMetric]
    rule: "Rule"

    def __init__(self, rule):
        self.metrics = OrderedDict()
        self.rule = rule

    # @contract(name='tuple,seq(string)', total='float|int', incremental=Sequence)
    def set_metric(
        self,
        name: Tuple[str, ...],
        total: Union[float, int],
        title: Optional[str] = None,
        description: Optional[str] = None,
        incremental: Optional[SampledSequence] = None,
        cumulative: Optional[SampledSequence] = None,
    ):
        check_isinstance(name, tuple)
        self.metrics[name] = EvaluatedMetric(
            total=total, incremental=incremental, title=title, description=description, cumulative=cumulative,
        )

    def __repr__(self):
        return "RuleEvaluationResult(%s, %s)" % (self.rule, self.metrics)


class Rule(metaclass=ABCMeta):
    @abstractmethod
    def evaluate(self, context: RuleEvaluationContext, result: RuleEvaluationResult):
        """ Evaluates the rule in this context.

            Must make at least one call to

                result.set_violation()
        """


def evaluate_rules(
    poses_sequence, interval: SampledSequence[Timestamp], world: PlacedObject, ego_name: str,
) -> Dict[str, RuleEvaluationResult]:
    from duckietown_world.world_duckietown import create_lane_highlight

    lane_pose_seq = create_lane_highlight(poses_sequence, world)
    from duckietown_world.rules import DeviationFromCenterLine
    from duckietown_world.rules import InDrivableLane
    from duckietown_world.rules import DeviationHeading
    from duckietown_world.rules import DrivenLength
    from duckietown_world.rules import DrivenLengthConsecutive
    from duckietown_world.rules import SurvivalTime

    rules = OrderedDict()
    rules["deviation-heading"] = DeviationHeading()
    rules["in-drivable-lane"] = InDrivableLane()
    rules["deviation-center-line"] = DeviationFromCenterLine()
    rules["driving-distance"] = DrivenLength()
    rules["driving-distance-consecutive"] = DrivenLengthConsecutive()
    rules["survival_time"] = SurvivalTime()

    context = RuleEvaluationContext(
        interval=interval,
        world=world,
        ego_name=ego_name,
        lane_pose_seq=lane_pose_seq,
        pose_seq=poses_sequence,
    )

    evaluated = OrderedDict()
    for name, rule in rules.items():
        result = RuleEvaluationResult(rule)
        rule.evaluate(context, result)
        evaluated[name] = result
    return evaluated


def make_timeseries(evaluated) -> Dict[str, "TimeseriesPlot"]:
    timeseries = OrderedDict()
    for k, rer in evaluated.items():
        from duckietown_world.rules import RuleEvaluationResult
        from duckietown_world.svg_drawing.misc import TimeseriesPlot

        assert isinstance(rer, RuleEvaluationResult)

        for km, evaluated_metric in rer.metrics.items():
            assert isinstance(evaluated_metric, EvaluatedMetric)
            sequences = {}
            if evaluated_metric.incremental:
                sequences["incremental"] = evaluated_metric.incremental
            if evaluated_metric.cumulative:
                sequences["cumulative"] = evaluated_metric.cumulative

            kk = "/".join((k,) + km)
            # title = evaluated_metric.title + ( '(%s)' % evaluated_metric.title if km else "")
            title = evaluated_metric.title
            timeseries[kk] = TimeseriesPlot(title or kk, evaluated_metric.description, sequences)

    return timeseries



# coding=utf-8
import os

import numpy as np
from comptests import comptest, run_module_tests, get_comptests_output_dir

from duckietown_world.geo import PlacedObject, SE2Transform, get_meausurements_graph
from duckietown_world.seqs import Constant
from duckietown_world.utils.gvgen_ac import ACGvGen
from duckietown_world.world_duckietown.map_loading import load_map


@comptest
def m1():
    outdir = get_comptests_output_dir()
    gm = load_map("udem1")

    # dw = DuckietownWorld()
    # for map_name, tm in gym_maps.items():
    #     DW.root.set_object(map_name, tm)

    root = PlacedObject()

    world = PlacedObject()
    root.set_object("world", world)
    origin = SE2Transform([1, 10], np.deg2rad(10))
    world.set_object("tile_map", gm, ground_truth=Constant[SE2Transform](origin))

    # d = dw.as_json_dict()
    # print(json.dumps(d, indent=4))
    # print(yaml.safe_dump(d, default_flow_style=False))
    #

    G = get_meausurements_graph(root)
    fn = os.path.join(outdir, "out1.pdf")
    plot_measurement_graph(root, G, fn)
    #
    # world = ('world',)
    # for node in G.nodes():
    #     paths = list(all_simple_paths(G, world, node))
    #     paths = [transform_from_path(G, _) for _ in paths]
    #     if paths:
    #         mpath = paths[0]
    #         path = squash_path(mpath)
    #         print(path)
    #         # frozen = [_.at(t=0) for _ in path]
    #         # print(frozen)


class NoMeasurements(Exception):
    pass


def squash_one(p, preference):
    for m in p:
        if m.sr_type in preference:
            return m.transform

    msg = "Missing measurement of type %r" % ", ".join(preference)
    msg += "\n\nFound:\n%r" % "\n".join(_.sr_type.__repr__() for _ in p)
    raise NoMeasurements(msg)


def squash_path(path, preference=("ground_truth",)):
    res = []
    for p in path:
        res.append(squash_one(p, preference))
    return res


def transform_from_path(G, path):
    transforms = []
    for u, v in pairwise(list(path)):
        t = []
        for k, edge_data in G.get_edge_data(u, v).items():
            attr_dict = edge_data["attr_dict"]
            transform = attr_dict["sr"]
            t.append(transform)
        transforms.append(t)
    return transforms


import itertools


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


def plot_measurement_graph(root, G, out):
    d = os.path.dirname(out)
    if not os.path.exists(d):
        os.makedirs(d)
    gg = ACGvGen()

    node2item = {}
    for node in G.nodes():
        label = node[-1] if node else "root"
        ob = root.get_object_from_fqn(node)
        label += "\n%s" % ob
        item = gg.newItem(label)
        node2item[node] = item
    for u, v in G.edges():
        for k, data in G.get_edge_data(u, v).items():
            attr_dict = data["attr_dict"]
            transform = attr_dict["sr"]
            label = str(transform)
            src = node2item[u]
            dst = node2item[v]
            gg.newLink(src=src, dst=dst, label=label)

    fn = out + ".dot"
    with open(fn, "w") as f:
        f.write(gg.dot2())

    cmd = "dot -o%s -Tpdf %s" % (out, fn)
    os.system(cmd)
    print("Written to %s" % out)


if __name__ == "__main__":
    run_module_tests()



# coding=utf-8
import numpy as np
from comptests import comptest, run_module_tests
from numpy.testing import assert_almost_equal

import geometry as geo
from duckietown_world.world_duckietown import Integrator2D, GenericKinematicsSE2
from duckietown_world.world_duckietown.differential_drive_dynamics import (
    DifferentialDriveDynamicsParameters,
    WheelVelocityCommands,
)


@comptest
def integrator2D_test1():
    q0 = geo.SE2_from_translation_angle([0, 0], 0)
    v0 = geo.se2.zero()
    c0 = q0, v0
    s0 = Integrator2D.initialize(c0)

    commands = [1, 2]
    dt = 2
    s1 = s0.integrate(dt, commands)

    q1, v1 = s1.TSE2_from_state()


@comptest
def kinematics2d_test():
    q0 = geo.SE2_from_translation_angle([0, 0], 0)
    v0 = geo.se2.zero()
    c0 = q0, v0
    s0 = GenericKinematicsSE2.initialize(c0)

    k = 0.4
    radius = 1.0 / k
    print("radius: %s" % radius)
    v = 3.1
    perimeter = 2 * np.pi * radius
    dt_loop = perimeter / v

    w = 2 * np.pi / dt_loop

    dt = dt_loop * 0.25
    commands = geo.se2_from_linear_angular([v, 0], w)

    s1 = s0.integrate(dt, commands)
    s2 = s1.integrate(dt, commands)
    s3 = s2.integrate(dt, commands)
    s4 = s3.integrate(dt, commands)
    seq = [s0, s1, s2, s3, s4]
    for _ in seq:
        q1, v1 = _.TSE2_from_state()
        print("%s" % geo.SE2.friendly(q1))

    assert_almost_equal(geo.translation_from_SE2(s1.TSE2_from_state()[0]), [radius, radius])
    assert_almost_equal(geo.translation_from_SE2(s2.TSE2_from_state()[0]), [0, radius * 2])
    assert_almost_equal(geo.translation_from_SE2(s3.TSE2_from_state()[0]), [-radius, radius])
    assert_almost_equal(geo.translation_from_SE2(s4.TSE2_from_state()[0]), [0, 0])


@comptest
def dd_test():
    radius = 0.1
    radius_left = radius
    radius_right = radius
    wheel_distance = 0.5
    dddp = DifferentialDriveDynamicsParameters(
        radius_left=radius_left, radius_right=radius_right, wheel_distance=wheel_distance,
    )
    q0 = geo.SE2_from_translation_angle([0, 0], 0)
    v0 = geo.se2.zero()
    c0 = q0, v0
    s0 = dddp.initialize(c0)

    omega = 0.3
    commands = WheelVelocityCommands(omega, omega)
    dt = 4.0
    s1 = s0.integrate(dt, commands)

    q1, v1 = s1.TSE2_from_state()
    print(geo.se2.friendly(v1))
    print(geo.SE2.friendly(q1))

    p1, theta = geo.translation_angle_from_SE2(q1)
    assert_almost_equal(p1, [dt * radius * omega, 0])

    omega_left = 0.3
    omega_right = 0
    commands = WheelVelocityCommands(omega_left, omega_right)
    dt = 4.0
    s1 = s0.integrate(dt, commands)

    q1, v1 = s1.TSE2_from_state()
    p1, theta = geo.translation_angle_from_SE2(q1)
    print(geo.se2.friendly(v1))
    print(geo.SE2.friendly(q1))

    # TODO: finish these tests
    # assert_almost_equal(p1[0], [dt * radius * omega_left / 2, 0])


if __name__ == "__main__":
    run_module_tests()



# coding=utf-8
from dataclasses import dataclass
from typing import Any, Dict, Optional

from duckietown_world import SampledSequence, SE2Transform
from ..world_duckietown import DuckietownMap

__all__ = [
    "RobotTrajectories",
    "SimulatorLog",
]


@dataclass
class RobotTrajectories:
    pose: SampledSequence[SE2Transform]
    # wheels_velocities: SampledSequence
    # actions: SampledSequence
    velocity: SampledSequence
    observations: SampledSequence
    commands: SampledSequence


@dataclass
class SimulatorLog:
    duckietown: DuckietownMap

    render_time: Optional[Any]
    robots: Dict[str, RobotTrajectories]
